from typing import Dict, List, Optional, Tuple, Union

import numpy as np

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import functools
import multiprocessing
import os
from collections import ChainMap

import cv2
import pandas as pd
from numba import int64, jit, njit, prange, uint8
from shapely.geometry import Polygon

from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists, check_if_valid_img,
                                check_instance, check_int, check_str,
                                check_valid_array, check_valid_lst)
from simba.utils.enums import Defaults, Formats, GeometryEnum, Options
from simba.utils.errors import ArrayError, FrameRangeError, InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_core_cnt,
                                    find_files_of_filetypes_in_directory,
                                    get_fn_ext, get_video_meta_data,
                                    read_frm_of_video)


class ImageMixin(object):
    """
    Methods to slice and compute attributes of images and comparing those attributes across sequential images.

    This can be helpful when the behaviors studied are very subtle and the signal is very low in relation to the
    noise within the pose-estimated data. In these use-cases, we cannot use pose-estimated data directly, and we instead
    study histograms, contours and other image metrics within images derived from the intersection of geometries
    (like a circle around the nose) across sequential images. Often these methods are called using image masks created from
    pose-estimated points within ``simba.mixins.geometry_mixin.GeometryMixin`` methods.

    .. important::
        If there is non-pose related noise in the environment (e.g., there are non-experiment related light sources that goes on and off, or other image noise that doesn't necesserily affect pose-estimation reliability),
        this will negatively affect the reliability of most image attribute comparisons.

    """

    def __init__(self):
        pass

    @staticmethod
    def brightness_intensity(
        imgs: List[np.ndarray], ignore_black: Optional[bool] = True
    ) -> List[float]:
        """
        Compute the average brightness intensity within each image within a list.

        :param List[np.ndarray] imgs: List of images as arrays to calculate average brightness intensity within.
        :param Optional[bool] ignore_black: If True, ignores black pixels. If the images are sliced non-rectangular geometric shapes created by ``slice_shapes_in_img``, then pixels that don't belong to the shape has been masked in black.
        :returns List[float]: List of floats of size len(imgs) with brightness intensities.

        :example:
        >>> img = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/0.png').astype(np.uint8)
        >>> ImageMixin.brightness_intensity(imgs=[img], ignore_black=False)
        >>> [159.0]
        """
        results = []
        check_instance(
            source=f"{ImageMixin().brightness_intensity.__name__} imgs",
            instance=imgs,
            accepted_types=list,
        )
        for cnt, img in enumerate(imgs):
            check_instance(
                source=f"{ImageMixin().brightness_intensity.__name__} img {cnt}",
                instance=img,
                accepted_types=np.ndarray,
            )
            if len(img) == 0:
                results.append(0)
            else:
                if ignore_black:
                    results.append(np.ceil(np.average(img[img != 0])))
                else:
                    results.append(np.ceil(np.average(img)))
        return results

    @staticmethod
    def get_histocomparison(
        img_1: np.ndarray,
        img_2: np.ndarray,
        method: Optional[
            Literal[
                "chi_square",
                "correlation",
                "intersection",
                "bhattacharyya",
                "hellinger",
                "chi_square_alternative",
                "kl_divergence",
            ]
        ] = "correlation",
        absolute: Optional[bool] = True,
    ):
        """
        :example:
        >>> img_1 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/0.png').astype(np.uint8)
        >>> img_2 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/3.png').astype(np.uint8)
        >>> ImageMixin.get_histocomparison(img_1=img_1, img_2=img_2, method='chi_square_alternative')
        """
        check_if_valid_img(
            source=f"{ImageMixin.get_histocomparison.__name__} img_1", data=img_1
        )
        check_if_valid_img(
            source=f"{ImageMixin.get_histocomparison.__name__} img_2", data=img_2
        )
        check_str(
            name=f"{ImageMixin().get_histocomparison.__name__} method",
            value=method,
            options=list(GeometryEnum.HISTOGRAM_COMPARISON_MAP.value.keys()),
        )
        method = GeometryEnum.HISTOGRAM_COMPARISON_MAP.value[method]
        if absolute:
            return abs(
                cv2.compareHist(
                    img_1.astype(np.float32), img_2.astype(np.float32), method
                )
            )
        else:
            return cv2.compareHist(
                img_1.astype(np.float32), img_2.astype(np.float32), method
            )

    @staticmethod
    def get_contourmatch(
        img_1: np.ndarray,
        img_2: np.ndarray,
        mode: Optional[Literal["all", "exterior"]] = "all",
        method: Optional[Literal["simple", "none", "l2", "kcos"]] = "simple",
        canny: Optional[bool] = True,
    ) -> float:
        """
        Calculate contour similarity between two images.

        :param np.ndarray img_1: First input image (numpy array).
        :param np.ndarray img_2: Second input image (numpy array).
        :param Optional[Literal['all', 'exterior']] method: Method for contour extraction. Options: 'all' (all contours) or 'exterior' (only exterior contours). Defaults to 'all'.
        :return float: Contour similarity score between the two images. Lower values indicate greater similarity, and higher values indicate greater dissimilarity.

        :example:
        >>> img_1 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/0.png').astype(np.uint8)
        >>> img_2 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/3.png').astype(np.uint8)
        >>> ImageMixin.get_contourmatch(img_1=img_1, img_2=img_2, method='exterior')
        """

        check_if_valid_img(
            source=f"{ImageMixin.get_contourmatch.__name__} img_1", data=img_1
        )
        check_if_valid_img(
            source=f"{ImageMixin.get_contourmatch.__name__} img_2", data=img_2
        )
        check_str(
            name=f"{ImageMixin().get_contourmatch.__name__} mode",
            value=mode,
            options=list(GeometryEnum.CONTOURS_MODE_MAP.value.keys()),
        )
        check_str(
            name=f"{ImageMixin.find_contours.__name__} method",
            value=method,
            options=list(GeometryEnum.CONTOURS_RETRIEVAL_MAP.value.keys()),
        )
        if canny:
            img_1 = ImageMixin().canny_edge_detection(img=img_1)
            img_2 = ImageMixin().canny_edge_detection(img=img_2)
        img_1_contours = ImageMixin().find_contours(img=img_1, mode=mode, method=method)
        img_2_contours = ImageMixin().find_contours(img=img_2, mode=mode, method=method)
        return cv2.matchShapes(
            img_1_contours[0], img_2_contours[0], cv2.CONTOURS_MATCH_I1, 0.0
        )

    @staticmethod
    def slice_shapes_in_img(
        img: Union[np.ndarray, Tuple[cv2.VideoCapture, int]],
        geometries: List[Union[Polygon, np.ndarray]],
    ) -> List[np.ndarray]:
        """
        Slice regions of interest (ROIs) from an image based on provided shapes.

        .. note::
           Use for slicing one or several static geometries from a single image. If you have several images,
           and shifting geometries across images, consider ``simba.mixins.image_mixin.ImageMixin.slice_shapes_in_imgs``
           which uses CPU multiprocessing.

        :param Union[np.ndarray, Tuple[cv2.VideoCapture, int]] img: Either an image in numpy array format OR a tuple with cv2.VideoCapture object and the frame index.
        :param List[Union[Polygon, np.ndarray]] img: A list of shapes either as vertices in a numpy array, or as shapely Polygons.
        :returns List[np.ndarray]: List of sliced ROIs from the input image.

        >>> img = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/videos/img_comparisons_4/1.png')
        >>> img_video = cv2.VideoCapture('/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/videos/Example_1.mp4')
        >>> data_path = '/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/csv/outlier_corrected_movement_location/Example_1.csv'
        >>> data = pd.read_csv(data_path, nrows=4, usecols=['Nose_x', 'Nose_y']).fillna(-1).values.astype(np.int64)
        >>> shapes = []
        >>> for frm_data in data: shapes.append(GeometryMixin().bodyparts_to_circle(frm_data, 100))
        >>> ImageMixin().slice_shapes_in_img(img=(img_video, 1), shapes=shapes)
        """

        result = []
        check_instance(
            source=f"{ImageMixin().slice_shapes_in_img.__name__} img",
            instance=img,
            accepted_types=(tuple, np.ndarray),
        )
        check_instance(
            source=f"{ImageMixin().slice_shapes_in_img.__name__} shapes",
            instance=geometries,
            accepted_types=list,
        )
        for shape_cnt, shape in enumerate(geometries):
            check_instance(
                source=f"{ImageMixin().slice_shapes_in_img.__name__} shapes {shape_cnt}",
                instance=shape,
                accepted_types=(Polygon, np.ndarray),
            )
        if isinstance(img, tuple):
            check_instance(
                source=f"{ImageMixin().slice_shapes_in_img.__name__} img tuple first entry",
                instance=img[0],
                accepted_types=cv2.VideoCapture,
            )
            frm_cnt = int(img[0].get(cv2.CAP_PROP_FRAME_COUNT))
            check_int(
                name=f"{ImageMixin().slice_shapes_in_img.__name__} video frame count",
                value=img[1],
                max_value=frm_cnt,
                min_value=0,
            )
            img[0].set(1, img[1])
            _, img = img[0].read()
        corrected_shapes = []
        for shape in geometries:
            if isinstance(shape, np.ndarray):
                shape[shape < 0] = 0
                corrected_shapes.append(shape)
            else:
                shape = np.array(shape.exterior.coords).astype(np.int64)
                shape[shape < 0] = 0
                corrected_shapes.append(shape)
        shapes = corrected_shapes
        del corrected_shapes
        for shape_cnt, shape in enumerate(shapes):
            x, y, w, h = cv2.boundingRect(shape)
            roi_img = img[y : y + h, x : x + w].copy()
            mask = np.zeros_like(roi_img, np.uint8)
            cv2.drawContours(
                mask, [shape - (x, y)], -1, (255, 255, 255), -1, cv2.LINE_AA
            )
            result.append(cv2.bitwise_and(roi_img, mask))
        return result

    @staticmethod
    def canny_edge_detection(
        img: np.ndarray,
        threshold_1: int = 30,
        threshold_2: int = 200,
        aperture_size: int = 3,
        l2_gradient: bool = False,
    ) -> np.ndarray:
        """
        Apply Canny edge detection to the input image.
        """
        check_if_valid_img(source=f"{ImageMixin.img_moments.__name__}", data=img)
        check_int(
            name=f"{ImageMixin.img_moments.__name__} threshold_1",
            value=threshold_1,
            min_value=1,
        )
        check_int(
            name=f"{ImageMixin.img_moments.__name__} threshold_2",
            value=threshold_2,
            min_value=1,
        )
        check_int(
            name=f"{ImageMixin.img_moments.__name__} aperture_size",
            value=aperture_size,
            min_value=1,
        )
        if len(img.shape) >= 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        return cv2.Canny(
            img,
            threshold1=threshold_1,
            threshold2=threshold_2,
            apertureSize=aperture_size,
            L2gradient=l2_gradient,
        )

    @staticmethod
    def img_moments(img: np.ndarray, hu_moments: Optional[bool] = False) -> np.ndarray:
        """
        Compute image moments.

        :param np.ndarray img: The input image.
        :param  Optional[bool] img: If True, returns the 7 Hu moments. Else, returns the moments.
        :returns np.ndarray: A 24x1 2d-array if hu_moments is False, 7x1 2d-array if hu_moments is True.

        :example:
        >>> img_1 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/0.png').astype(np.uint8)
        >>> ImageMixin.img_moments(img=img_1, hu_moments=True)
        >>> [[ 1.01270313e-03], [ 8.85983106e-10], [ 4.67680675e-13], [ 1.00442018e-12], [-4.64181508e-25], [-2.49036749e-17], [ 5.08375216e-25]]
        """
        check_if_valid_img(source=f"{ImageMixin.img_moments.__name__}", data=img)
        if len(img.shape) >= 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        if not hu_moments:
            return np.array(list(cv2.moments(img).values())).reshape(-1, 1)
        else:
            return cv2.HuMoments(cv2.moments(img))

    @staticmethod
    def find_contours(
        img: np.ndarray,
        mode: Optional[Literal["all", "exterior"]] = "all",
        method: Optional[Literal["simple", "none", "l1", "kcos"]] = "simple",
    ) -> np.ndarray:
        """
        Find contours in the input image.

        :param np.ndarray img: Input image as a NumPy array.
        :param Optional[Literal['all', 'exterior']] img: Contour retrieval mode. E.g., which contours should be kept. Default is 'all'.
        :param Optional[Literal['simple', 'none', 'l1', 'kcos']]: Contour approximation method. Default is 'simple'.
        """

        check_if_valid_img(source=f"{ImageMixin.find_contours.__name__} img", data=img)
        check_str(
            name=f"{ImageMixin.find_contours.__name__} mode",
            value=mode,
            options=list(GeometryEnum.CONTOURS_MODE_MAP.value.keys()),
        )
        check_str(
            name=f"{ImageMixin.find_contours.__name__} method",
            value=method,
            options=list(GeometryEnum.CONTOURS_RETRIEVAL_MAP.value.keys()),
        )
        mode = GeometryEnum.CONTOURS_MODE_MAP.value[mode]
        method = GeometryEnum.CONTOURS_RETRIEVAL_MAP.value[method]
        if len(img.shape) >= 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        if mode in [0, 1]:
            return cv2.findContours(img, mode, method)[1]
        else:
            cnts, hierarchy = cv2.findContours(img, mode, method)[-2:]  # TODO
            interior_contours = []
            for i in range(len(cnts)):
                if (
                    hierarchy[0][i][3] == -1
                ):  # Contour with no parent (interior contour)
                    interior_contours.append(cnts[i])

    @staticmethod
    def orb_matching_similarity_(
        img_1: np.ndarray,
        img_2: np.ndarray,
        method: Literal["knn", "match", "radius"] = "knn",
        mask: Optional[np.ndarray] = None,
        threshold: Optional[int] = 0.75,
    ) -> int:
        """Perform ORB feature matching between two sets of images.

        >>> img_1 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/0.png').astype(np.uint8)
        >>> img_2 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/10.png').astype(np.uint8)
        >>> ImageMixin().orb_matching_similarity_(img_1=img_1, img_2=img_2, method='radius')
        >>> 4
        """

        kp1, des1 = cv2.ORB_create().detectAndCompute(img_1, mask)
        kp2, des2 = cv2.ORB_create().detectAndCompute(img_2, mask)
        sliced_matches = None
        if method == "knn":
            matches = cv2.BFMatcher().knnMatch(des1, des2, k=2)
            sliced_matches = [
                m for m, n in matches if m.distance < threshold * n.distance
            ]
        if method == "match":
            matches = cv2.BFMatcher(cv2.NORM_HAMMING, crossCheck=True).match(des1, des2)
            sliced_matches = [match for match in matches if match.distance <= threshold]
        if method == "radius":
            matches = cv2.BFMatcher().radiusMatch(des1, des2, maxDistance=threshold)
            sliced_matches = [item for sublist in matches for item in sublist]
        return len(sliced_matches)

    @staticmethod
    def _template_matching_cpu_helper(
        data: np.ndarray, video_path: Union[str, os.PathLike], target_frm: np.ndarray
    ):
        """Helper called from ``simba.mixins.image_mixin.ImageMixins.template_matching_cpu()``"""
        cap = cv2.VideoCapture(video_path)
        start, end, current = data[0], data[-1], data[0]
        cap.set(1, start)
        results = {}
        while current < end:
            print(f"Processing frame {current}...")
            _, img = cap.read()
            result = cv2.matchTemplate(img, target_frm, cv2.TM_CCOEFF_NORMED)
            _, _, _, max_loc = cv2.minMaxLoc(result)
            results[current] = {"p": np.max(result), "loc": max_loc}
            current += 1
        return results

    @staticmethod
    def template_matching_cpu(
        video_path: Union[str, os.PathLike],
        img: np.ndarray,
        core_cnt: Optional[int] = -1,
        return_img: Optional[bool] = False,
    ) -> Tuple[int, dict, Union[None, np.ndarray]]:
        """
        Perform template matching on CPU using multiprocessing for parallelization.

        E.g., having a cropped image, find the image and frame number in a video it most likely has been cropped from.

        :param Union[str, os.PathLike] video_path: Path to the video file on disk.
        :param np.ndarray img: Template image for matching. E.g., a cropped image from ``video_path``.
        :param Optional[int] core_cnt: Number of CPU cores to use for parallel processing. Default is -1 (max available cores).
        :param Optional[bool] return_img: Whether to return the annotated best match image with rectangle around matched template area. Default is False.
        :returns Tuple[ int, dict, Union[None, np.ndarray]]: A tuple containing: (i) int: frame index of the frame with the best match.
                 (ii) dict: Dictionary containing results (probability and match location) for each frame.
                 (iii) Union[None, np.ndarray]: Annotated image with rectangles around matches (if return_img is True), otherwise None.

        :example:
        >>> img = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/videos/Screenshot 2024-01-17 at 12.45.55 PM.png')
        >>> results = ImageMixin().template_matching_cpu(video_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/videos/Together_1.avi', img=img, return_img=True)
        """

        results, found_img = [], None
        check_if_valid_img(data=img)
        if core_cnt == -1:
            core_cnt = find_core_cnt()[0]
        frame_cnt = get_video_meta_data(video_path=video_path)["frame_count"]
        frm_idx = np.arange(0, frame_cnt + 1)
        chunk_size = len(frm_idx) // core_cnt
        remainder = len(frm_idx) % core_cnt
        split_frm_idx = [
            frm_idx[
                i * chunk_size
                + min(i, remainder) : (i + 1) * chunk_size
                + min(i + 1, remainder)
            ]
            for i in range(core_cnt)
        ]
        with multiprocessing.Pool(
            core_cnt, maxtasksperchild=Defaults.LARGE_MAX_TASK_PER_CHILD.value
        ) as pool:
            constants = functools.partial(
                ImageMixin()._template_matching_cpu_helper,
                video_path=video_path,
                target_frm=img,
            )
            for cnt, result in enumerate(
                pool.imap(constants, split_frm_idx, chunksize=1)
            ):
                results.append(result)
        pool.terminate()
        pool.join()
        results = dict(ChainMap(*results))

        max_value, max_frm = -np.inf, None
        for k, v in results.items():
            if v["p"] > max_value:
                max_value = v["p"]
                max_frm = k

        if return_img:
            h, w, _ = img.shape
            found_img = read_frm_of_video(video_path=video_path, frame_index=max_frm)
            loc = results[max_frm]["loc"]
            found_img = cv2.rectangle(
                found_img,
                (int(loc[0]), int(loc[1])),
                (int(loc[0]) + w, int(loc[1] + h)),
                (0, 255, 0),
                2,
            )
        return max_frm, results, found_img

    def template_matching_gpu(self):
        # TODO
        pass

    @staticmethod
    def img_to_bw(
        img: np.ndarray,
        lower_thresh: Optional[int] = 20,
        upper_thresh: Optional[int] = 250,
        invert: Optional[bool] = True,
    ) -> np.ndarray:
        """
        Convert an image to black and white (binary).

        :param np.ndarray img: Input image as a NumPy array.
        :param Optional[int] lower_thresh: Lower threshold value for binary conversion. Pixels below this value become black. Default is 20.
        :param Optional[int] upper_thresh: Upper threshold value for binary conversion. Pixels above this value become white. Default is 250.
        :param Optional[bool] invert: Flag indicating whether to invert the binary image (black becomes white and vice versa). Default is True.
        :return np.ndarray: Binary black and white image.
        """
        check_if_valid_img(
            data=img, source=ImageMixin().segment_img_horizontal.__name__
        )
        check_int(
            name=f"{ImageMixin().segment_img_horizontal.__name__} lower_thresh",
            value=lower_thresh,
            max_value=255,
            min_value=1,
        )
        check_int(
            name=f"{ImageMixin().segment_img_horizontal.__name__} upper_thresh",
            value=upper_thresh,
            max_value=255,
            min_value=1,
        )
        if len(img) > 2:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        if not invert:
            return cv2.threshold(img, lower_thresh, upper_thresh, cv2.THRESH_BINARY)[1]
        else:
            return ~cv2.threshold(img, lower_thresh, upper_thresh, cv2.THRESH_BINARY)[1]

    @staticmethod
    def segment_img_horizontal(
        img: np.ndarray,
        pct: int,
        lower: Optional[bool] = True,
        both: Optional[bool] = False,
    ) -> np.ndarray:
        """
        Segment a horizontal part of the input image.

        This function segments either the lower, upper, or both lower and upper part of the input image based on the specified percentage.

        :param np.ndarray img: Input image as a NumPy array.
        :param int pct: Percentage of the image to be segmented. If `lower` is True, it represents the lower part; if False, it represents the upper part.
        :param Optional[bool] lower: Flag indicating whether to segment the lower part (True) or upper part (False) of the image. Default is True.
        :return np.array: Segmented part of the image.
        """

        check_if_valid_img(
            data=img, source=ImageMixin().segment_img_horizontal.__name__
        )
        check_int(
            name=f"{ImageMixin().segment_img_horizontal.__name__} pct",
            value=pct,
            min_value=1,
            max_value=99,
        )
        sliced_height = int(img.shape[0] * pct / 100)
        if both:
            return img[sliced_height : img.shape[0] - sliced_height, :]
        elif lower:
            return img[img.shape[0] - sliced_height :, :]
        else:
            return img[:sliced_height, :]

    @staticmethod
    def segment_img_vertical(
        img: np.ndarray,
        pct: int,
        left: Optional[bool] = True,
        both: Optional[bool] = False,
    ) -> np.ndarray:
        """
        Segment a vertical part of the input image.

        This function segments either the left, right or both the left and right part of  input image based on the specified percentage.

        :param np.ndarray img: Input image as a NumPy array.
        :param int pct: Percentage of the image to be segmented. If `lower` is True, it represents the lower part; if False, it represents the upper part.
        :param Optional[bool] lower: Flag indicating whether to segment the lower part (True) or upper part (False) of the image. Default is True.
        :return np.array: Segmented part of the image.
        """

        check_if_valid_img(data=img, source=ImageMixin().segment_img_vertical.__name__)
        check_int(
            name=f"{ImageMixin().segment_img_vertical.__name__} pct",
            value=pct,
            min_value=1,
            max_value=99,
        )
        sliced_width = int(img.shape[1] * pct / 100)
        if both:
            return img[:, sliced_width : img.shape[1] - sliced_width]
        elif left:
            return img[:, :sliced_width]
        else:
            return img[:, img.shape[1] - sliced_width :]

    @staticmethod
    def add_img_border_and_flood_fill(
        img: np.array, invert: Optional[bool] = False, size: Optional[int] = 1
    ) -> np.ndarray:
        """
        Add a border to the input image and perform flood fill.

        E.g., Used to remove any black pixel areas connected to the border of the image. Used to remove noise
        if noise is defined as being connected to the edges of the image.

        .. image:: _static/img/add_img_border_and_flood_fill.png
           :width: 400
           :align: center

        :param np.ndarray img: Input image as a numpy array.
        :param Optional[bool] invert: If false, make black border and floodfill black pixels with white. If True, make white border and floodfill white pixels with black. Default False.
        :param Optional[bool] size: Size of border. Default 1 pixel.
        """

        check_if_valid_img(
            data=img, source=ImageMixin().add_img_border_and_flood_fill.__name__
        )
        check_int(
            name=f"{ImageMixin().add_img_border_and_flood_fill.__name__} size",
            value=size,
            min_value=1,
        )
        if len(img.shape) > 2:
            raise InvalidInputError(
                msg="Floodfill requires 2d image",
                source=ImageMixin().add_img_border_and_flood_fill.__name__,
            )
        if not invert:
            img = cv2.copyMakeBorder(
                img, size, size, size, size, cv2.BORDER_CONSTANT, value=0
            )
            mask = np.zeros((img.shape[0] + 2, img.shape[1] + 2), dtype=np.uint8)
            img = cv2.floodFill(
                img, mask=mask, seedPoint=(0, 0), newVal=(255, 255, 255)
            )[1]

        else:
            img = cv2.copyMakeBorder(
                img, size, size, size, size, cv2.BORDER_CONSTANT, value=255
            )
            mask = np.zeros((img.shape[0] + 2, img.shape[1] + 2), dtype=np.uint8)
            img = cv2.floodFill(img, mask=mask, seedPoint=(0, 0), newVal=(0, 0, 0))[1]

        return img[size:-size, size:-size]

    @staticmethod
    def _image_reader_helper(img_paths: List[str]):
        """Multiprocessing helper for ``ImageMixin().read_all_img_in_dir``"""
        results = {}
        for img_path in img_paths:
            results[get_fn_ext(filepath=img_path)[1]] = cv2.imread(img_path)
        return results

    @staticmethod
    def read_all_img_in_dir(
        dir: Union[str, os.PathLike], core_cnt: Optional[int] = -1
    ) -> Dict[str, np.ndarray]:
        """
        Helper to read in all images within a directory using multiprocessing.
        Returns a dictionary with the image name as key and the images in array format as values.

        :example:
        >>> imgs = ImageMixin().read_all_img_in_dir(dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/Together_4_cropped_frames')
        """
        check_if_dir_exists(in_dir=dir)
        file_paths = find_files_of_filetypes_in_directory(
            directory=dir,
            extensions=list(Options.ALL_IMAGE_FORMAT_OPTIONS.value),
            raise_error=True,
        )
        if core_cnt == -1:
            core_cnt = find_core_cnt()[0]
        chunk_size = len(file_paths) // core_cnt
        file_paths = [
            file_paths[
                i * chunk_size
                + min(i, len(file_paths) % core_cnt) : (i + 1) * chunk_size
                + min(i + 1, len(file_paths) % core_cnt)
            ]
            for i in range(core_cnt)
        ]
        imgs = {}
        with multiprocessing.Pool(
            core_cnt, maxtasksperchild=Defaults.LARGE_MAX_TASK_PER_CHILD.value
        ) as pool:
            for cnt, result in enumerate(
                pool.imap(ImageMixin()._image_reader_helper, file_paths, chunksize=1)
            ):
                imgs.update(result)
        pool.join()
        pool.terminate()
        return imgs

    @staticmethod
    @njit([(uint8[:, :, :, :], uint8[:, :, :, :]), (uint8[:, :, :], uint8[:, :, :])])
    def img_stack_mse(imgs_1: np.ndarray, imgs_2: np.ndarray) -> np.ndarray:
        """
        Pairwise comparison of images in two stacks of equal length using mean squared errors.

        .. note::
           Images has to be in uint8 format.

        :param np.ndarray imgs_1: First three (non-color) or four (color) dimensional stack of images in array format.
        :param np.ndarray imgs_1: Second three (non-color) or four (color) dimensional stack of images in array format.
        :return np.ndarray: Array of size len(imgs_1) comparing ``imgs_1`` and ``imgs_2`` at each index using mean squared errors at each pixel location.

        :example:
        >>> img_1 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/0.png').astype(np.uint8)
        >>> img_2 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/10.png').astype(np.uint8)
        >>> imgs_1 = np.stack((img_1, img_2)); imgs_2 = np.stack((img_2, img_2))
        >>> ImageMixin.img_stack_mse(imgs_1=imgs_1, imgs_2=imgs_2)
        >>> [637,   0]
        >>> imgs = ImageMixin().read_all_img_in_dir(dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/Together_4_cropped_frames')
        >>> imgs_1 = np.stack(imgs.values())
        >>> imgs_2 = np.roll(imgs_1,-1, axis=0)
        >>> mse = ImageMixin().img_stack_mse(imgs_1=imgs_1, imgs_2=imgs_1)
        """

        results = np.full((imgs_1.shape[0]), np.nan)
        for i in range(imgs_1.shape[0]):
            results[i] = np.sum((imgs_1[i] - imgs_2[i]) ** 2) / float(
                imgs_1[i].shape[0] * imgs_2[i].shape[1]
            )
        return results.astype(np.int64)

    @staticmethod
    @njit([(uint8[:, :, :, :], int64), (uint8[:, :, :], int64)])
    def img_sliding_mse(imgs: np.ndarray, slide_size: int = 1) -> np.ndarray:
        """Pairwise comparison of images in sliding windows using mean squared errors

        :example:
        >>> imgs = ImageMixin().read_all_img_in_dir(dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/Together_4_cropped_frames')
        >>> imgs = np.stack(imgs.values())
        >>> mse = ImageMixin().img_sliding_mse(imgs=imgs, slide_size=2)
        """

        results = np.full((imgs.shape[0]), 0)
        for i in prange(slide_size, imgs.shape[0]):
            results[i] = np.sum((imgs[i - slide_size] - imgs[i]) ** 2) / float(
                imgs[i - slide_size].shape[0] * imgs[i].shape[1]
            )
        return results.astype(int64)

    @staticmethod
    def _read_img_batch_from_video_helper(
        frm_idx: np.ndarray, video_path: Union[str, os.PathLike]
    ):
        """Multiprocess helper used by read_img_batch_from_video to read in images from video file."""
        start_idx, end_frm, current_frm = frm_idx[0], frm_idx[-1] + 1, frm_idx[0]
        results = {}
        cap = cv2.VideoCapture(video_path)
        cap.set(1, current_frm)
        while current_frm < end_frm:
            results[current_frm] = cap.read()[1]
            current_frm += 1
        return results

    @staticmethod
    def read_img_batch_from_video(
        video_path: Union[str, os.PathLike],
        start_frm: int,
        end_frm: int,
        core_cnt: Optional[int] = -1,
    ) -> Dict[int, np.ndarray]:
        """
        Read a batch of frames from a video file. This method reads frames from a specified range of frames within a video file using multiprocessing.

        :param Union[str, os.PathLike] video_path: Path to the video file.
        :param int start_frm: Starting frame index.
        :param int end_frm: Ending frame index.
        :param Optionalint] core_cnt: Number of CPU cores to use for parallel processing. Default is -1, indicating using all available cores.
        :param Optional[bool] greyscale: If True, reads the images as greyscale. If False, then as original color scale. Default: True.
        :returns Dict[int, np.ndarray]: A dictionary containing frame indices as keys and corresponding frame arrays as values.

        :example:
        >>> ImageMixin().read_img_batch_from_video(video_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/videos/Together_1.avi', start_frm=0, end_frm=50)
        """
        check_file_exist_and_readable(file_path=video_path)
        video_meta_data = get_video_meta_data(video_path=video_path)
        check_int(
            name=ImageMixin().__class__.__name__,
            value=start_frm,
            min_value=0,
            max_value=video_meta_data["frame_count"],
        )
        check_int(
            name=ImageMixin().__class__.__name__,
            value=end_frm,
            min_value=0,
            max_value=video_meta_data["frame_count"],
        )
        check_int(name=ImageMixin().__class__.__name__, value=core_cnt, min_value=-1)
        if core_cnt < 0:
            core_cnt = multiprocessing.cpu_count()
        if end_frm <= start_frm:
            FrameRangeError(
                msg=f"Start frame ({start_frm}) has to be before end frame ({end_frm})",
                source=ImageMixin().__class__.__name__,
            )
        frm_lst = np.array_split(np.arange(start_frm, end_frm + 1), core_cnt)
        results = {}
        with multiprocessing.Pool(
            core_cnt, maxtasksperchild=Defaults.LARGE_MAX_TASK_PER_CHILD.value
        ) as pool:
            constants = functools.partial(
                ImageMixin()._read_img_batch_from_video_helper, video_path=video_path
            )
            for cnt, result in enumerate(pool.imap(constants, frm_lst, chunksize=1)):
                results.update(result)
        pool.join()
        pool.terminate()
        return results

    @staticmethod
    def img_emd(
        imgs: List[np.ndarray] = None,
        img_1: Optional[np.ndarray] = None,
        img_2: Optional[np.ndarray] = None,
        lower_bound: Optional[float] = 0.5,
    ):
        """
        Compute Wasserstein distance between two images represented as numpy arrays.

        :example:
        >>> img_1 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/videos/Example_1_frames/24.png', 0).astype(np.float32)
        >>> img_2 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/videos/Example_1_frames/1984.png', 0).astype(np.float32)
        >>> img_emd(img_1=img_1, img_2=img_3, lower_bound=0.5)
        >>> 10.658767700195312
        """
        check_float(
            name=f"{ImageMixin.img_emd.__name__} lower bound",
            min_value=0.0,
            value=lower_bound,
        )
        if (img_1 is None and img_1 is None and imgs is None) or (
            img_1 is not None and img_2 is None
        ):
            raise InvalidInputError(
                msg="Pass img_1 and img_2 OR imgs", source=ImageMixin.__class__.__name__
            )
        if (img_1 is not None) and (img_2 is not None):
            check_if_valid_img(data=img_1, source=ImageMixin.img_emd.__name__)
            check_if_valid_img(data=img_2, source=ImageMixin.img_emd.__name__)
            if img_1.ndim > 2:
                img_1 = cv2.cvtColor(img_1, cv2.COLOR_BGR2GRAY)
            if img_2.ndim > 2:
                img_2 = cv2.cvtColor(img_2, cv2.COLOR_BGR2GRAY)
        else:
            check_valid_lst(
                data=imgs,
                source=ImageMixin.img_emd.__name__,
                valid_dtypes=(np.ndarray,),
                exact_len=2,
            )
            check_if_valid_img(data=imgs[0], source=ImageMixin.img_emd.__name__)
            check_if_valid_img(data=imgs[1], source=ImageMixin.img_emd.__name__)
            if imgs[0].ndim > 2:
                img_1 = cv2.cvtColor(imgs[0], cv2.COLOR_BGR2GRAY)
            if imgs[1].ndim > 2:
                img_2 = cv2.cvtColor(imgs[1], cv2.COLOR_BGR2GRAY)
        return cv2.EMD(
            img_1.astype(np.float32),
            img_2.astype(np.float32),
            cv2.DIST_L2,
            lowerBound=lower_bound,
        )[0]

    @staticmethod
    @jit(nopython=True)
    def img_matrix_mse(imgs: np.ndarray) -> np.ndarray:
        """
        Jitted compute mean squared error matrix table for a stack of images.

        :example:
        >>> imgs = ImageMixin().read_img_batch_from_video(video_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/videos/Together_1.avi', start_frm=0, end_frm=50)
        >>> imgs = np.stack(list(imgs.values()))
        >>> ImageMixin().img_matrix_mse(imgs=imgs)
        """
        results = np.full((imgs.shape[0], imgs.shape[0]), 0.0)
        for i in prange(imgs.shape[0]):
            for j in range(i + 1, imgs.shape[0]):
                val = np.sum((imgs[i] - imgs[j]) ** 2) / float(
                    imgs[i].shape[0] * imgs[j].shape[1]
                )
                results[i, j] = val
                results[j, i] = val
        return results.astype(np.int32)

    @staticmethod
    @njit("(uint8[:, :, :, :],)", fastmath=True)
    def img_stack_to_greyscale(imgs: np.ndarray):
        """
        Jitted conversion of a 4D stack of color images (RGB format) to grayscale.

        :parameter np.ndarray imgs: A 4D array representing color images. It should have the shape (num_images, height, width, 3) where the last dimension represents the color channels (R, G, B).
        :returns np.ndarray: A 3D array containing the grayscale versions of the input images. The shape of the output array is (num_images, height, width).

        :example:
        >>> imgs = ImageMixin().read_img_batch_from_video( video_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/videos/Together_1.avi', start_frm=0, end_frm=100)
        >>> imgs = np.stack(list(imgs.values()))
        >>> imgs_gray = img_stack_to_greyscale(imgs=imgs)
        """
        results = np.full((imgs.shape[0], imgs.shape[1], imgs.shape[2]), np.nan).astype(
            np.uint8
        )
        for i in prange(imgs.shape[0]):
            vals = (
                0.07 * imgs[i][:, :, 2]
                + 0.72 * imgs[i][:, :, 1]
                + 0.21 * imgs[i][:, :, 0]
            )
            results[i] = vals.astype(np.uint8)
        return results

    @staticmethod
    def _slice_shapes_in_imgs_array_helper(
        data: Tuple[np.ndarray, np.ndarray]
    ) -> List[np.ndarray]:
        """
        Private multiprocess helper called from ``simba.mixins.image_mixin.ImageMixins.slice_shapes_in_imgs()`` to slice shapes from
        an array of images.
        """
        img, in_shapes = data[0], data[1]
        shapes, results = [], []
        for shape in in_shapes:
            shape = np.array(shape.exterior.coords).astype(np.int64)
            shape[shape < 0] = 0
            shapes.append(shape)
        for shape_cnt, shape in enumerate(shapes):
            x, y, w, h = cv2.boundingRect(shape)
            roi_img = img[y : y + h, x : x + w].copy()
            mask = np.zeros_like(roi_img, np.uint8)
            cv2.drawContours(
                mask, [shape - (x, y)], -1, (255, 255, 255), -1, cv2.LINE_AA
            )
            results.append(cv2.bitwise_and(roi_img, mask))
        return results

    @staticmethod
    def pad_img_stack(
        image_dict: Dict[int, np.ndarray], pad_value: Optional[int] = 0
    ) -> Dict[int, np.ndarray]:
        """
        Pad images in a dictionary stack to have the same dimensions (the same dimension is represented by the largest image in the stack)

        :param Dict[int, np.ndarray] image_dict: A dictionary mapping integer keys to numpy arrays representing images.
        :param Optional[int] pad_value: The value used for padding. Defaults to 0 (black)
        :return Dict[int, np.ndarray]: A dictionary mapping integer keys to numpy arrays representing padded images.
        """
        check_instance(
            source=ImageMixin.pad_img_stack.__name__,
            instance=image_dict,
            accepted_types=(dict,),
        )
        check_int(
            name=f"{ImageMixin.pad_img_stack.__name__} pad_value",
            value=pad_value,
            max_value=255,
            min_value=0,
        )
        max_height = max(image.shape[0] for image in image_dict.values())
        max_width = max(image.shape[1] for image in image_dict.values())
        padded_images = {}
        for key, image in image_dict.items():
            pad_height = max_height - image.shape[0]
            pad_width = max_width - image.shape[1]
            padded_image = np.pad(
                image,
                ((0, pad_height), (0, pad_width), (0, 0)),
                mode="constant",
                constant_values=pad_value,
            )
            padded_images[key] = padded_image
        return padded_images

    @staticmethod
    def img_stack_to_video(
        imgs: Dict[int, np.ndarray],
        save_path: Union[str, os.PathLike],
        fps: int,
        verbose: Optional[bool] = True,
    ):

        check_instance(
            source=ImageMixin.img_stack_to_video.__name__,
            instance=imgs,
            accepted_types=(dict,),
        )
        img_sizes = set()
        for k, v in imgs.items():
            img_sizes.add(v.shape)
        if len(list(img_sizes)) > 1:
            imgs = ImageMixin.pad_img_stack(imgs)
        imgs = np.stack(imgs.values())
        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        writer = cv2.VideoWriter(
            save_path, fourcc, fps, (imgs[0].shape[1], imgs[0].shape[0])
        )
        for i in range(imgs.shape[0]):
            if verbose:
                print(f"Writing img {i + 1}...")
            writer.write(imgs[i])
        writer.release()

    @staticmethod
    def _slice_shapes_in_video_file_helper(
        data: np.ndarray, video_path: Union[str, os.PathLike], verbose: bool
    ):
        cap = cv2.VideoCapture(video_path)
        start_frm, current_frm, end_frm = data[0][0], data[0][0], data[-1][0]
        cap.set(1, start_frm)
        results = {}
        idx_cnt = 0
        while current_frm <= end_frm:
            if verbose:
                print(f"Processing frame {current_frm}...")
            img = cap.read(current_frm)[1].astype(np.uint8)
            shape = np.array(data[idx_cnt][1].exterior.coords).astype(np.int64)
            shape[shape < 0] = 0
            x, y, w, h = cv2.boundingRect(shape)
            roi_img = img[y : y + h, x : x + w].copy()
            mask = np.zeros_like(roi_img, np.uint8)
            cv2.drawContours(
                mask, [shape - (x, y)], -1, (255, 255, 255), -1, cv2.LINE_AA
            )
            results[current_frm] = cv2.bitwise_and(roi_img, mask).astype(np.uint8)
            current_frm += 1
            idx_cnt += 1
        return results

    def slice_shapes_in_imgs(
        self,
        imgs: Union[np.ndarray, os.PathLike],
        shapes: Union[np.ndarray, List[Polygon]],
        core_cnt: Optional[int] = -1,
        verbose: Optional[bool] = False,
    ) -> List[np.ndarray]:
        """
        Slice regions from a stack of images or a video file, where the regions are based on defined shapes. Uses multiprocessing.

        For example, given a stack of N images, and N*X geometries representing the region around the animal body-part(s),
        slice out the X geometries from each of the N images and return the sliced areas.

        .. image:: _static/img/slice_shapes_in_imgs_4.gif
           :width: 400
           :align: center

        .. image:: _static/img/bodyparts_to_circle_1.gif
           :width: 400
           :align: center

        .. image:: _static/img/bodyparts_to_circle_2.gif
           :width: 400
           :align: center

        :example I:
        >>> imgs = ImageMixin().read_img_batch_from_video( video_path='/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/videos/Example_1.mp4', start_frm=0, end_frm=10)
        >>> imgs = np.stack(list(imgs.values()))
        >>> imgs_gray = ImageMixin().img_stack_to_greyscale(imgs=imgs)
        >>> data = pd.read_csv('/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/csv/outlier_corrected_movement_location/Example_1.csv', nrows=11).fillna(-1)
        >>> nose_array, tail_array = data.loc[0:10, ['Nose_x', 'Nose_y']].values.astype(np.float32), data.loc[0:10, ['Tail_base_x', 'Tail_base_y']].values.astype(np.float32)
        >>> nose_shapes, tail_shapes = [], []
        >>> for frm_data in nose_array: nose_shapes.append(GeometryMixin().bodyparts_to_circle(frm_data, 80))
        >>> for frm_data in tail_array: tail_shapes.append(GeometryMixin().bodyparts_to_circle(frm_data, 80))
        >>> shapes = np.array(np.vstack([nose_shapes, tail_shapes]).T)
        >>> sliced_images = ImageMixin().slice_shapes_in_imgs(imgs=imgs_gray, shapes=shapes)

        :example II:
        >>> video_path = '/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/videos/Example_1_clipped.mp4'
        >>> data_path = r'/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/csv/outlier_corrected_movement_location/Example_1_clipped.csv'
        >>> df = pd.read_csv(data_path, usecols=['Nose_x', 'Nose_y', 'Tail_base_x', 'Tail_base_y']).fillna(0).values.astype(int)
        >>> data = df.reshape(len(df), -1, int(df.shape[1]/2))
        >>> geometries = GeometryMixin().multiframe_bodyparts_to_line(data=data, buffer=30, px_per_mm=4.1)
        >>> imgs = ImageMixin().slice_shapes_in_imgs(imgs=video_path, shapes=geometries)


        """
        timer = SimbaTimer(start=True)
        check_int(
            name=f"{ImageMixin().slice_shapes_in_imgs.__name__} core count",
            value=core_cnt,
            min_value=-1,
        )
        if core_cnt == -1:
            core_cnt = find_core_cnt()[0]
        check_instance(
            source=ImageMixin().slice_shapes_in_imgs.__name__,
            instance=imgs,
            accepted_types=(np.ndarray, str),
        )
        check_instance(
            source=ImageMixin().slice_shapes_in_imgs.__name__,
            instance=shapes,
            accepted_types=(np.ndarray, list),
        )
        if isinstance(shapes, np.ndarray):
            check_valid_array(
                data=shapes,
                source=ImageMixin().slice_shapes_in_imgs.__name__,
                accepted_ndims=(2,),
                accepted_dtypes=[Polygon],
            )
        else:
            check_valid_lst(
                data=shapes,
                source=ImageMixin().slice_shapes_in_imgs.__name__,
                valid_dtypes=[Polygon],
            )
            shapes = np.array(shapes)
        if isinstance(imgs, np.ndarray):
            check_valid_array(
                data=imgs,
                source=ImageMixin().slice_shapes_in_imgs.__name__,
                accepted_ndims=(4, 3),
            )
            if shapes.shape[0] != imgs.shape[0]:
                raise ArrayError(
                    msg=f"The image array ({imgs.shape[0]}) and shapes array ({shapes.shape[0]}) have unequal length.",
                    source=ImageMixin().slice_shapes_in_imgs.__name__,
                )
        else:
            check_file_exist_and_readable(file_path=imgs)
            video_meta_data = get_video_meta_data(video_path=imgs)
            if shapes.shape[0] != video_meta_data["frame_count"]:
                raise ArrayError(
                    msg=f'The image array ({video_meta_data["frame_count"]}) and shapes array ({shapes.shape[0]}) have unequal length.',
                    source=ImageMixin().slice_shapes_in_imgs.__name__,
                )
        results = []
        if isinstance(imgs, np.ndarray):
            with multiprocessing.Pool(
                core_cnt, maxtasksperchild=Defaults.LARGE_MAX_TASK_PER_CHILD.value
            ) as pool:
                for cnt, result in enumerate(
                    pool.imap(
                        self._slice_shapes_in_imgs_array_helper,
                        zip(imgs, shapes),
                        chunksize=1,
                    )
                ):
                    results.append(result)
        else:
            shapes = np.array_split(
                np.column_stack((np.arange(len(shapes)), shapes)), core_cnt
            )
            with multiprocessing.Pool(
                core_cnt, maxtasksperchild=Defaults.LARGE_MAX_TASK_PER_CHILD.value
            ) as pool:
                constants = functools.partial(
                    self._slice_shapes_in_video_file_helper,
                    video_path=imgs,
                    verbose=verbose,
                )
                for cnt, result in enumerate(pool.imap(constants, shapes, chunksize=1)):
                    results.append(result)
                results = dict(ChainMap(*results))
        pool.join()
        pool.terminate()
        timer.stop_timer()
        stdout_success(
            msg="Geometry image slicing complete.",
            elapsed_time=timer.elapsed_time_str,
            source=self.__class__.__name__,
        )
        return results


# imgs = ImageMixin().read_img_batch_from_video( video_path='/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/videos/Example_1.mp4', start_frm=0, end_frm=10)
# imgs = np.stack(list(imgs.values()))
# imgs_gray = ImageMixin().img_stack_to_greyscale(imgs=imgs)
# data = pd.read_csv('/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/csv/outlier_corrected_movement_location/Example_1.csv', nrows=11).fillna(-1)
# nose_array, tail_array = data.loc[0:10, ['Nose_x', 'Nose_y']].values.astype(np.float32), data.loc[0:10, ['Tail_base_x', 'Tail_base_y']].values.astype(np.float32)
# nose_shapes, tail_shapes = [], []
# from simba.mixins.geometry_mixin import GeometryMixin
# for frm_data in nose_array: nose_shapes.append(GeometryMixin().bodyparts_to_circle(frm_data, 80))
# for frm_data in tail_array: tail_shapes.append(GeometryMixin().bodyparts_to_circle(frm_data, 80))
# shapes = np.array(np.vstack([nose_shapes, tail_shapes]).T)
# sliced_images = ImageMixin().slice_shapes_in_imgs(imgs=imgs_gray, shapes=shapes)

# imgs_.shape
# imgs = ImageMixin().read_all_img_in_dir(dir='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/Together_4_cropped_frames')
# imgs_1 = np.stack(imgs.values())
# mse = ImageMixin().img_sliding_mse(imgs=imgs_1, slide_size=2)
# #imgs_2 = np.roll(imgs_1,-1, axis=0)
# #mse = ImageMixin().img_mse(imgs_1=imgs_1, imgs_2=imgs_1)
# mse = ImageMixin().img_sliding_mse(imgs=imgs_1, slide_size=2)

# img_1 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/0.png').astype(np.uint8)
# img_2 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/10.png').astype(np.uint8)
#
# img_1 = cv2.cvtColor(img_1, cv2.COLOR_BGR2GRAY)
# img_2 = cv2.cvtColor(img_2, cv2.COLOR_BGR2GRAY)
#
# imgs_1 = np.stack((img_1, img_2))
# imgs_2 = np.stack((img_2, img_2))
# ImageMixin.img_mse(imgs_1=imgs_1, imgs_2=imgs_2)

# ImageMixin.img_mse_test(imgs_1=imgs_1)


# res = ImageMixin.img_moments(img=img_1, hu_moments=True)
# img_1 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/0.png').astype(np.uint8)
# img_2 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/10.png').astype(np.uint8)
# ImageMixin.get_contourmatch(img_1=img_1, img_2=img_2, mode='exterior')


# img = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/videos/img_comparisons_4/1.png')
# img_video = cv2.VideoCapture('/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/videos/Example_1.mp4')
# data_path = '/Users/simon/Desktop/envs/troubleshooting/Emergence/project_folder/csv/outlier_corrected_movement_location/Example_1.csv'
# data = pd.read_csv(data_path, nrows=4, usecols=['Nose_x', 'Nose_y']).fillna(-1).values.astype(np.int64)
# shapes = []
# for frm_data in data: shapes.append(GeometryMixin().bodyparts_to_circle(frm_data, 100))
#
# ImageMixin().slice_shapes_in_img(img=(img_video, 1), shapes=shapes)


# img_2 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/3.png').astype(np.uint8)
# ImageMixin.get_contourmatch(img_1=img_1, img_2=img_2, method='exterior')

# img = cv2.VideoCapture('/Users/simon/Desktop/envs/simba_dev/tests/data/test_projects/zebrafish/project_folder/videos/20200730_AB_7dpf_850nm_0003.mp4')
# ImageMixin.slice_shapes_in_img(img=(img, 1))


# img_1 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/0.png').astype(np.uint8)
# img_2 = cv2.imread('/Users/simon/Desktop/envs/troubleshooting/khan/project_folder/videos/stitched_frames/1.png').astype(np.uint8)
# ImageMixin.get_contourmatch(img_1=img_1, img_2=img_2, method='all', canny=True)
#



import glob
import os
from copy import deepcopy

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.third_party_label_appenders.tools import (
    check_stop_events_prior_to_start_events, fix_uneven_start_stop_count,
    read_bento_files, read_boris_annotation_files, read_deepethogram_files,
    read_ethovision_files, read_observer_files, read_solomon_files)
from simba.utils.checks import (check_if_dir_exists,
                                check_if_filepath_list_is_empty)
from simba.utils.enums import Methods
from simba.utils.errors import (
    ThirdPartyAnnotationEventCountError, ThirdPartyAnnotationFileNotFoundError,
    ThirdPartyAnnotationOverlapError, ThirdPartyAnnotationsAdditionalClfError,
    ThirdPartyAnnotationsMissingAnnotationsError,
    ThirdPartyAnnotationsOutsidePoseEstimationDataError)
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df
from simba.utils.warnings import (
    ThirdPartyAnnotationEventCountWarning,
    ThirdPartyAnnotationFileNotFoundWarning,
    ThirdPartyAnnotationOverlapWarning,
    ThirdPartyAnnotationsAdditionalClfWarning,
    ThirdPartyAnnotationsMissingAnnotationsWarning,
    ThirdPartyAnnotationsOutsidePoseEstimationDataWarning)

BORIS = "BORIS"
DEEPETHOGRAM = "DEEPETHOGRAM"
ETHOVISION = "ETHOVISION"
OBSERVER = "OBSERVER"
SOLOMON = "SOLOMON"
BENTO = "BENTO"
BEHAVIOR = "BEHAVIOR"


class ThirdPartyLabelAppender(ConfigReader):
    """
    Concatenate third-party annotations to featurized pose-estimation datasets in SimBA.

    :param str app: Third-party application. OPTIONS: ['BORIS', 'BENTO', 'DEEPETHOGRAM', 'ETHOVISION', 'OBSERVER', 'SOLOMON'].
    :param str config_path: path to SimBA project config file in Configparser format.
    :param str data_dir: Directory holding third-party annotation data files.
    :param dict settings: User-defined settings including how to handle errors, logging, and data file types associated with the third-party application.

    ... note::
       `Third-party import tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot.md>`__.

       `BENTO: expected input <https://github.com/sgoldenlab/simba/blob/master/misc/bento_example.annot`__.
       `BORIS: expected input <https://github.com/sgoldenlab/simba/blob/master/misc/boris_example.csv>`__..
       `DEEPETHOGRAM: expected input <https://github.com/sgoldenlab/simba/blob/master/misc/deep_ethogram_labels.csv>`__.
       `ETHOVISION: expected input <https://github.com/sgoldenlab/simba/blob/master/misc/ethovision_example.xlsx>`__.
       `OBSERVER: expected input I <https://github.com/sgoldenlab/simba/blob/master/misc/Observer_example_1.xlsx>`__.
       `OBSERVER: expected input II <https://github.com/sgoldenlab/simba/blob/master/misc/Observer_example_2.xlsx>`__.
       `SOLOMON: expected input II <https://github.com/sgoldenlab/simba/blob/master/misc/solomon_example.csv>`__.


    Examples
    ----------
    >>> settings = {'log': True,  'file_format': 'csv', 'errors': {'INVALID annotations file data format': 'WARNING',
    >>>                                                             'ADDITIONAL third-party behavior detected': 'NONE',
    >>>                                                             'Annotations EVENT COUNT conflict': 'WARNING',
    >>>                                                             'Annotations OVERLAP inaccuracy': 'WARNING',
    >>>                                                             'ZERO third-party video behavior annotations found': 'WARNING',
    >>>                                                             'Annotations and pose FRAME COUNT conflict': 'WARNING',
    >>>                                                             'Annotations data file NOT FOUND': 'WARNING'}}
    >>> third_party_importer = ThirdPartyLabelAppender(app='BORIS', config_path=r'MySimBAConfigPath', data_dir=r'MyDeepEthogramDir', settings=settings)
    >>> third_party_importer.run()

    References
    ----------

    .. [1] `DeepEthogram repo <https://github.com/jbohnslav/deepethogram>`__.
    .. [2]  Segalin et al., eLife, https://doi.org/10.7554/eLife.63720
    .. [3] `Behavioral Observation Research Interactive Software (BORIS) user guide <https://boris.readthedocs.io/en/latest/#>`__.
    .. [4] `Noldus Ethovision XT <https://www.noldus.com/ethovision-xt>`__.
    .. [5] `Noldus Observer XT <https://www.noldus.com/observer-xt>`__.
    .. [6] `Solomon coder user-guide (PDF) <https://solomon.andraspeter.com/Solomon%20Intro.pdf>`__.

    """

    def __init__(self, app: str, config_path: str, data_dir: str, settings: dict):
        super().__init__(config_path=config_path)
        check_if_dir_exists(in_dir=data_dir)
        self.data_dir, self.app, self.settings = data_dir, app, settings
        self.err_settings = self.settings["errors"]
        if self.settings["log"]:
            self.create_logger(
                path=os.path.join(self.logs_path, f"BORIS_append_{self.datetime}.log")
            )
        self.data_file_paths = glob.glob(
            self.data_dir + f'/*.{self.settings["file_format"]}'
        )
        self.data_file_paths = [x for x in self.data_file_paths if "~$" not in x]
        check_if_filepath_list_is_empty(
            filepaths=self.data_file_paths,
            error_msg=f'SIMBA ERROR: ZERO {app} {self.settings["file_format"]} files found in {data_dir} directory',
        )
        check_if_filepath_list_is_empty(
            filepaths=self.feature_file_paths,
            error_msg="SIMBA ERROR: ZERO files found in the project_folder/csv/features_extracted directory",
        )
        print(f"Processing {str(len(self.feature_file_paths))} {app} file(s)...")

    def __check_annotation_clf_df_integrity(self, df=pd.DataFrame):
        clf_name = df["BEHAVIOR"].loc[0]
        if len(df[df["EVENT"] == "START"]) != len(df[df["EVENT"] == "STOP"]):
            if (
                self.err_settings[Methods.THIRD_PARTY_EVENT_COUNT_CONFLICT.value]
                == Methods.WARNING.value
            ):
                ThirdPartyAnnotationEventCountWarning(
                    video_name=self.video_name,
                    clf_name=clf_name,
                    start_event_cnt=len(df[df["EVENT"] == "START"]),
                    stop_event_cnt=len(df[df["EVENT"] == "STOP"]),
                    log_status=self.settings["log"],
                )
                df = fix_uneven_start_stop_count(data=df)

            elif (
                self.err_settings[Methods.THIRD_PARTY_EVENT_COUNT_CONFLICT.value]
                == Methods.ERROR.value
            ):
                raise ThirdPartyAnnotationEventCountError(
                    video_name=self.video_name,
                    clf_name=clf_name,
                    start_event_cnt=len(df[df["EVENT"] == "START"]),
                    stop_event_cnt=len(df[df["EVENT"] == "STOP"]),
                )

        else:
            start = df["FRAME"][df["EVENT"] == "START"].reset_index(drop=True)
            stop = df["FRAME"][df["EVENT"] == "STOP"].reset_index(drop=True)
            df = pd.concat([start, stop], axis=1)
            df.columns = ["START", "STOP"]

        overlaps_idx = check_stop_events_prior_to_start_events(df=df)
        if overlaps_idx:
            if (
                self.err_settings[Methods.THIRD_PARTY_EVENT_OVERLAP.value]
                == Methods.WARNING.value
            ):
                ThirdPartyAnnotationOverlapWarning(
                    video_name=self.video_name,
                    clf_name=clf_name,
                    log_status=self.settings["log"],
                )
                df = df.drop(index=overlaps_idx).reset_index(drop=True)

            elif (
                self.err_settings[Methods.THIRD_PARTY_EVENT_OVERLAP.value]
                == Methods.ERROR.value
            ):
                raise ThirdPartyAnnotationOverlapError(
                    video_name=self.video_name, clf_name=clf_name
                )

        return df

    def run(self):
        data = None
        print(
            f"Reading in {str(len(self.data_file_paths))} {self.app} annotation files..."
        )
        if self.app == BORIS:
            data = read_boris_annotation_files(
                data_paths=self.data_file_paths,
                error_setting=self.settings["errors"][
                    Methods.INVALID_THIRD_PARTY_APPENDER_FILE.value
                ],
                video_info_df=self.video_info_df,
                log_setting=self.settings["log"],
            )
        elif self.app == DEEPETHOGRAM:
            data = read_deepethogram_files(
                data_paths=self.data_file_paths,
                error_setting=self.settings["errors"][
                    Methods.INVALID_THIRD_PARTY_APPENDER_FILE.value
                ],
                log_setting=self.settings["log"],
            )
        elif self.app == ETHOVISION:
            data = read_ethovision_files(
                data_paths=self.data_file_paths,
                error_setting=self.settings["errors"][
                    Methods.INVALID_THIRD_PARTY_APPENDER_FILE.value
                ],
                video_info_df=self.video_info_df,
                log_setting=self.settings["log"],
            )
        elif self.app == OBSERVER:
            data = read_observer_files(
                data_paths=self.data_file_paths,
                error_setting=self.settings["errors"][
                    Methods.INVALID_THIRD_PARTY_APPENDER_FILE.value
                ],
                video_info_df=self.video_info_df,
                log_setting=self.settings["log"],
            )

        elif self.app == SOLOMON:
            data = read_solomon_files(
                data_paths=self.data_file_paths,
                error_setting=self.settings["errors"][
                    Methods.INVALID_THIRD_PARTY_APPENDER_FILE.value
                ],
                video_info_df=self.video_info_df,
                log_setting=self.settings["log"],
            )

        elif self.app == BENTO:
            data = read_bento_files(
                data_paths=self.data_file_paths,
                error_setting=self.settings["errors"][
                    Methods.INVALID_THIRD_PARTY_APPENDER_FILE.value
                ],
                video_info_df=self.video_info_df,
                log_setting=self.settings["log"],
            )

        for file_cnt, file_path in enumerate(self.feature_file_paths):
            _, self.video_name, _ = get_fn_ext(filepath=file_path)
            print(f"Processing annotations for {self.video_name} video...")
            if self.video_name not in data.keys():
                if (
                    self.err_settings[
                        Methods.THIRD_PARTY_ANNOTATION_FILE_NOT_FOUND.value
                    ]
                    == Methods.WARNING.value
                ):
                    ThirdPartyAnnotationFileNotFoundWarning(
                        video_name=self.video_name, log_status=self.settings["log"]
                    )
                    continue
                if (
                    self.err_settings[
                        Methods.THIRD_PARTY_ANNOTATION_FILE_NOT_FOUND.value
                    ]
                    == Methods.ERROR.value
                ):
                    raise ThirdPartyAnnotationFileNotFoundError(
                        video_name=self.video_name
                    )
            annot_df = data[self.video_name].drop_duplicates().reset_index(drop=True)
            additional_clfs = list(
                set(annot_df[BEHAVIOR].unique()) - set(self.clf_names)
            )
            if (
                additional_clfs
                and self.err_settings[Methods.ADDITIONAL_THIRD_PARTY_CLFS.value]
                == Methods.WARNING.value
            ):
                ThirdPartyAnnotationsAdditionalClfWarning(
                    video_name=self.video_name,
                    clf_names=additional_clfs,
                    log_status=self.settings["log"],
                )
            elif (
                additional_clfs
                and self.err_settings[Methods.ADDITIONAL_THIRD_PARTY_CLFS.value]
                == Methods.ERROR.value
            ):
                raise ThirdPartyAnnotationsAdditionalClfError(
                    video_name=self.video_name, clf_names=additional_clfs
                )

            features_df = read_df(file_path=file_path, file_type=self.file_type)
            out_df = deepcopy(features_df)
            for clf in self.clf_names:
                clf_annot = annot_df[(annot_df[BEHAVIOR] == clf)].reset_index(drop=True)

                if len(clf_annot) == 0:
                    if (
                        self.err_settings[
                            Methods.ZERO_THIRD_PARTY_VIDEO_BEHAVIOR_ANNOTATIONS.value
                        ]
                        == Methods.WARNING.value
                    ):
                        ThirdPartyAnnotationsMissingAnnotationsWarning(
                            video_name=self.video_name,
                            clf_names=self.clf_names,
                            log_status=self.settings["log"],
                        )
                        out_df[clf] = 0
                        continue
                    elif (
                        self.err_settings[
                            Methods.ZERO_THIRD_PARTY_VIDEO_BEHAVIOR_ANNOTATIONS.value
                        ]
                        == Methods.ERROR.value
                    ):
                        raise ThirdPartyAnnotationsMissingAnnotationsError(
                            video_name=self.video_name, clf_names=self.clf_names
                        )
                clf_annot = self.__check_annotation_clf_df_integrity(df=clf_annot)
                annot_idx = list(
                    clf_annot.apply(
                        lambda x: list(range(int(x["START"]), int(x["STOP"]) + 1)), 1
                    )
                )
                annot_idx = [x for xs in annot_idx for x in xs]
                idx_diff = list(set(annot_idx) - set(out_df.index))
                if len(idx_diff) > 0:
                    if (
                        self.err_settings[
                            Methods.THIRD_PARTY_FRAME_COUNT_CONFLICT.value
                        ]
                        == Methods.WARNING.value
                    ):
                        ThirdPartyAnnotationsOutsidePoseEstimationDataWarning(
                            video_name=self.video_name,
                            clf_name=clf,
                            frm_cnt=out_df.index[-1],
                            first_error_frm=idx_diff[0],
                            ambiguous_cnt=len(idx_diff),
                            log_status=self.settings["log"],
                        )
                    elif (
                        self.err_settings[
                            Methods.THIRD_PARTY_FRAME_COUNT_CONFLICT.value
                        ]
                        == Methods.ERROR.value
                    ):
                        raise ThirdPartyAnnotationsOutsidePoseEstimationDataError(
                            video_name=self.video_name,
                            clf_name=clf,
                            frm_cnt=out_df.index[-1],
                            first_error_frm=idx_diff[0],
                            ambiguous_cnt=len(idx_diff),
                        )

                annot_idx = [x for x in annot_idx if x not in idx_diff]
                out_df[clf] = 0
                out_df.loc[annot_idx, clf] = 1
            save_path = os.path.join(
                self.targets_folder, self.video_name + "." + self.file_type
            )
            write_df(out_df, self.file_type, save_path)
            print(f"Saved {self.app} annotations for video {self.video_name}...")
        self.timer.stop_timer()
        stdout_success(
            msg=f"{self.app} annotations appended to dataset and saved in project_folder/csv/targets_inserted directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# settings = {'log': True,  'file_format': 'xlsx', 'errors': {'INVALID annotations file data format': 'WARNING',
#                                                            'ADDITIONAL third-party behavior detected': 'NONE',
#                                                            'Annotations EVENT COUNT conflict': 'WARNING',
#                                                            'Annotations OVERLAP inaccuracy': 'WARNING',
#                                                            'ZERO third-party video behavior annotations found': 'WARNING',
#                                                            'Annotations and pose FRAME COUNT conflict': 'WARNING',
#                                                            'Annotations data file NOT FOUND': 'WARNING'}}
# #
#
#
# test = ThirdPartyLabelAppender(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                data_dir='/Users/simon/Downloads/FIXED',
#                                settings=settings,
#                                app='BORIS')
# test.run()


# test = ThirdPartyLabelAppender(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                data_dir='/Users/simon/Desktop/envs/simba_dev/tests/test_data/deepethogram_example',
#                                settings=settings,
#                                app='DEEPETHOGRAM')
# test.run()


# test = ThirdPartyLabelAppender(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                data_dir=r'/Users/simon/Desktop/envs/simba_dev/tests/test_data/import_tests/ethovision_data',
#                                settings=settings,
#                                app='ETHOVISION')
# test.run()

# test = ThirdPartyLabelAppender(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                data_dir=r'/Users/simon/Desktop/envs/simba_dev/tests/test_data/solomon_import/solomon_import',
#                                settings=settings,
#                                app='SOLOMON')
# test.run()


# test = ThirdPartyLabelAppender(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                data_dir=r'/Users/simon/Desktop/envs/simba_dev/tests/test_data/bento_example',
#                                settings=settings,
#                                app='BENTO')
# test.run()



__author__ = "Simon Nilsson"

import os
from collections import deque
from copy import deepcopy
from typing import Dict, List, Optional

import cv2
import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.data import find_frame_numbers_from_time_stamp
from simba.utils.enums import Formats
from simba.utils.errors import FrameRangeError, NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class PathPlotterSingleCore(ConfigReader, PlottingMixin):
    """
    Create "path plots" videos and/or images detailing the movement paths of
    individual animals in SimBA.

    .. note::
        For improved run-time, see :meth:`simba.path_plotter_mp.PathPlotterMulticore` for multiprocess class.
       `Visualization tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. image:: _static/img/path_plot.png
       :width: 300
       :align: center

    :param str config_path: Path to SimBA project config file in Configparser format
    :param bool frame_setting: If True, individual frames will be created.
    :param bool video_setting: If True, compressed videos will be created.
    :param List[str] files_found: Data paths to create from which to create plots
    :param dict animal_attr: Animal body-parts and colors
    :param dict style_attr: Plot sttributes (line thickness, color, etc..)
    :param Optional[dict] slicing: If Dict, start time and end time of video slice to create path plot from. E.g., {'start_time': '00:00:01', 'end_time': '00:00:03'}. If None, creates path plot for entire video.



    Examples
    ----------
    >>> style_attr = {'width': 'As input', 'height': 'As input', 'line width': 5, 'font size': 5, 'font thickness': 2, 'circle size': 5, 'bg color': 'White', 'max lines': 100}
    >>> animal_attr = {0: ['Ear_right_1', 'Red']}
    >>> path_plotter = PathPlotterSingleCore(config_path=r'MyConfigPath', frame_setting=False, video_setting=True, style_attr=style_attr, animal_attr=animal_attr, files_found=['project_folder/csv/machine_results/MyVideo.csv']).run()
    """

    def __init__(
        self,
        config_path: str,
        frame_setting: bool,
        video_setting: bool,
        last_frame: bool,
        files_found: List[str],
        input_style_attr: dict,
        animal_attr: dict,
        input_clf_attr: dict,
        slicing: Optional[Dict] = None,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        (
            self.video_setting,
            self.frame_setting,
            self.input_style_attr,
            self.files_found,
            self.animal_attr,
            self.input_clf_attr,
            self.last_frame,
        ) = (
            video_setting,
            frame_setting,
            input_style_attr,
            files_found,
            animal_attr,
            input_clf_attr,
            last_frame,
        )
        if (not frame_setting) and (not video_setting) and (not last_frame):
            raise NoSpecifiedOutputError(
                msg="SIMBA ERROR: Please choice to create path frames and/or video path plots"
            )

        self.no_animals_path_plot, self.clf_attr, self.slicing = (
            len(animal_attr.keys()),
            None,
            slicing,
        )
        if not os.path.exists(self.path_plot_dir):
            os.makedirs(self.path_plot_dir)
        check_if_filepath_list_is_empty(
            filepaths=self.files_found,
            error_msg="Zero files found in the project_folder/csv/machine_results directory. To plot paths without performing machine classifications, use path plotter functions in [ROI] tab.",
        )
        print(f"Processing {str(len(self.files_found))} videos...")

    def run(self):
        """
        Method to create path plot videos and/or frames.Results are store in the
        'project_folder/frames/path_plots' directory of the SimBA project.
        """

        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, _, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.data_df = read_df(file_path, self.file_type).reset_index(drop=True)
            if self.slicing:
                frm_numbers = find_frame_numbers_from_time_stamp(
                    start_time=self.slicing["start_time"],
                    end_time=self.slicing["end_time"],
                    fps=self.fps,
                )
                if len(set(frm_numbers) - set(self.data_df.index)) > 0:
                    raise FrameRangeError(
                        msg=f'The chosen time-period ({self.slicing["start_time"]} - {self.slicing["end_time"]}) does not exist in {self.video_name}.'
                    )
                else:
                    self.data_df = self.data_df.loc[
                        frm_numbers[0] : frm_numbers[-1]
                    ].reset_index(drop=True)
            self.__get_styles()
            self.__get_deque_lookups()
            if self.video_setting:
                self.video_save_path = os.path.join(
                    self.path_plot_dir, self.video_name + ".mp4"
                )
                self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
                self.writer = cv2.VideoWriter(
                    self.video_save_path,
                    self.fourcc,
                    self.fps,
                    (self.style_attr["width"], self.style_attr["height"]),
                )

            if self.frame_setting:
                self.save_video_folder = os.path.join(
                    self.path_plot_dir, self.video_name
                )
                if not os.path.exists(self.save_video_folder):
                    os.makedirs(self.save_video_folder)

            if self.input_clf_attr:
                clf_names, self.clf_attr = [], {}
                self.clf_attr["attr"] = deepcopy(self.input_clf_attr)
                for v in self.input_clf_attr.values():
                    clf_names.append(v[0])
                self.clf_attr["data"] = self.data_df[clf_names]
                self.clf_attr["positions"] = self.data_df[
                    [self.animal_attr[0][0] + "_x", self.animal_attr[0][0] + "_y"]
                ]

            if self.last_frame:
                _ = self.make_path_plot(
                    data_df=self.data_df,
                    video_info=self.video_info,
                    style_attr=self.style_attr,
                    deque_dict=self.deque_dict,
                    clf_attr=self.clf_attr,
                    save_path=os.path.join(
                        self.path_plot_dir, self.video_name + "_final_frame.png"
                    ),
                )

            if self.video_setting or self.frame_setting:
                for frm_cnt in range(len(self.data_df)):
                    img = np.zeros(
                        (
                            int(self.video_info["Resolution_height"].values[0]),
                            int(self.video_info["Resolution_width"].values[0]),
                            3,
                        )
                    )
                    img[:] = self.style_attr["bg color"]
                    for animal_cnt, (animal_name, animal_data) in enumerate(
                        self.deque_dict.items()
                    ):
                        bp_x = int(
                            self.data_df.loc[
                                frm_cnt, "{}_{}".format(animal_data["bp"], "x")
                            ]
                        )
                        bp_y = int(
                            self.data_df.loc[
                                frm_cnt, "{}_{}".format(animal_data["bp"], "y")
                            ]
                        )
                        self.deque_dict[animal_name]["deque"].appendleft((bp_x, bp_y))
                    for animal_name, animal_data in self.deque_dict.items():
                        cv2.circle(
                            img,
                            (self.deque_dict[animal_name]["deque"][0]),
                            0,
                            self.deque_dict[animal_name]["clr"],
                            self.style_attr["circle size"],
                        )
                        cv2.putText(
                            img,
                            animal_name,
                            (self.deque_dict[animal_name]["deque"][0]),
                            Formats.FONT.value,
                            self.style_attr["font size"],
                            self.deque_dict[animal_name]["clr"],
                            self.style_attr["font thickness"],
                        )

                    for animal_name, animal_data in self.deque_dict.items():
                        for i in range(len(self.deque_dict[animal_name]["deque"]) - 1):
                            line_clr = self.deque_dict[animal_name]["clr"]
                            position_1 = self.deque_dict[animal_name]["deque"][i]
                            position_2 = self.deque_dict[animal_name]["deque"][i + 1]
                            cv2.line(
                                img,
                                position_1,
                                position_2,
                                line_clr,
                                self.style_attr["line width"],
                            )

                    if self.input_clf_attr:
                        animal_1_name = list(self.deque_dict.keys())[0]
                        animal_bp_x, animal_bp_y = (
                            self.deque_dict[animal_1_name]["bp"] + "_x",
                            self.deque_dict[animal_1_name]["bp"] + "_y",
                        )
                        for clf_cnt, clf_name in enumerate(
                            self.clf_attr["data"].columns
                        ):
                            clf_size = int(
                                self.clf_attr["attr"][clf_cnt][-1].split(": ")[-1]
                            )
                            clf_clr = self.color_dict[self.clf_attr["attr"][clf_cnt][1]]
                            sliced_df = self.clf_attr["data"].loc[0:frm_cnt]
                            sliced_df_idx = list(
                                sliced_df[
                                    sliced_df[self.clf_attr["attr"][clf_cnt][0]] == 1
                                ].index
                            )
                            locations = (
                                self.data_df.loc[
                                    sliced_df_idx, [animal_bp_x, animal_bp_y]
                                ]
                                .astype(int)
                                .values
                            )
                            for i in range(locations.shape[0]):
                                cv2.circle(
                                    img,
                                    (locations[i][0], locations[i][1]),
                                    0,
                                    clf_clr,
                                    clf_size,
                                )

                    img = cv2.resize(
                        img, (self.style_attr["width"], self.style_attr["height"])
                    )
                    if self.video_setting:
                        self.writer.write(np.uint8(img))
                    if self.frame_setting:
                        frm_name = os.path.join(
                            self.save_video_folder, str(frm_cnt) + ".png"
                        )
                        cv2.imwrite(frm_name, np.uint8(img))
                    print(
                        "Frame: {} / {}. Video: {} ({}/{})".format(
                            str(frm_cnt + 1),
                            str(len(self.data_df)),
                            self.video_name,
                            str(file_cnt + 1),
                            len(self.files_found),
                        )
                    )

                if self.video_setting:
                    self.writer.release()
                video_timer.stop_timer()
                print(
                    "Path visualization for video {} saved (elapsed time {}s)...".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )

        self.timer.stop_timer()
        stdout_success(
            msg=f"Path visualizations for {str(len(self.files_found))} videos saved in project_folder/frames/output/path_plots directory",
            elapsed_time=self.timer.elapsed_time_str,
        )

    def __get_styles(self):
        self.style_attr = {}
        if self.input_style_attr is not None:
            self.style_attr["bg color"] = self.color_dict[
                self.input_style_attr["bg color"]
            ]
            if self.input_style_attr["max lines"] == "entire video":
                self.style_attr["max lines"] = len(self.data_df)
            else:
                self.style_attr["max lines"] = int(
                    int(self.input_style_attr["max lines"])
                    * (int(self.video_info["fps"].values[0]) / 1000)
                )
            self.style_attr["font thickness"] = self.input_style_attr["font thickness"]
            self.style_attr["line width"] = self.input_style_attr["line width"]
            self.style_attr["font size"] = self.input_style_attr["font size"]
            self.style_attr["circle size"] = self.input_style_attr["circle size"]
            if self.input_style_attr["width"] == "As input":
                self.style_attr["width"], self.style_attr["height"] = int(
                    self.video_info["Resolution_width"].values[0]
                ), int(self.video_info["Resolution_height"].values[0])
            else:
                pass
        else:
            space_scaler, radius_scaler, res_scaler, font_scaler = 25, 10, 1500, 0.8
            self.style_attr["width"] = int(
                self.video_info["Resolution_width"].values[0]
            )
            self.style_attr["height"] = int(
                self.video_info["Resolution_height"].values[0]
            )
            max_res = max(self.style_attr["width"], self.style_attr["height"])
            self.style_attr["circle size"] = int(radius_scaler / (res_scaler / max_res))
            self.style_attr["font size"] = int(font_scaler / (res_scaler / max_res))
            self.style_attr["bg color"] = self.color_dict["White"]
            self.style_attr["max lines"] = len(self.data_df)
            self.style_attr["font thickness"] = 2
            self.style_attr["line width"] = 2

    def __get_deque_lookups(self):
        self.deque_dict = {}
        for animal_cnt, animal_data in self.animal_attr.items():
            animal_name = self.find_animal_name_from_body_part_name(
                bp_name=animal_data[0], bp_dict=self.animal_bp_dict
            )
            self.deque_dict[animal_name] = {}
            self.deque_dict[animal_name]["deque"] = deque(
                maxlen=self.style_attr["max lines"]
            )
            self.deque_dict[animal_name]["bp"] = self.animal_attr[animal_cnt][0]
            self.deque_dict[animal_name]["clr"] = self.color_dict[
                self.animal_attr[animal_cnt][1]
            ]


# style_attr = {'width': 'As input',
#               'height': 'As input',
#               'line width': 5,
#               'font size': 5,
#               'font thickness': 2,
#               'circle size': 5,
#               'bg color': 'White',
#               'max lines': 'entire video'}
#
# animal_attr = {0: ['Ear_right_1', 'Red'], 1: ['Ear_right_2', 'Green']}
# clf_attr = {0: ['Attack', 'Black', 'Size: 30'], 1: ['Sniffing', 'Black', 'Size: 30']}

# test = PathPlotterSingleCore(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                             frame_setting=False,
#                             video_setting=True,
#                              last_frame=True,
#                             input_style_attr=style_attr,
#                    animal_attr=animal_attr,
#                              input_clf_attr=clf_attr,
#                     files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv',
#                                  '/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'])
# test.run()

# test = PathPlotterSingleCore(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                              frame_setting=False,
#                              video_setting=True,
#                              last_frame=True,
#                              input_style_attr=style_attr,
#                              animal_attr=animal_attr,
#                              input_clf_attr=clf_attr,
#                              slicing = {'start_time': '00:00:01', 'end_time': '00:00:03'},
#                              files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'])
# test.run()



__author__ = "Simon Nilsson"

import numpy as np
from numba import jit, njit, prange
from numba.np.extensions import cross2d


@njit("(float32[:,:], int64[:], int64, int64)")
def process(S, P, a, b):
    signed_dist = cross2d(S[P] - S[a], S[b] - S[a])
    K = np.array(
        [i for s, i in zip(signed_dist, P) if s > 0 and i != a and i != b],
        dtype=np.int64,
    )
    if len(K) == 0:
        return [a, b]
    c = P[np.argmax(signed_dist)]
    return process(S, K, a, c)[:-1] + process(S, K, c, b)


@njit("(float32[:, :,:], types.unicode_type)", fastmath=True)
def jitted_hull(points: np.ndarray, target: str = "perimeter") -> np.ndarray:
    """
    Compute attributes (e.g., perimeter or area) of a polygon.

    :param array points: 3d array FRAMESxBODY-PARTxCOORDINATE
    :param str target: Options [perimeter, area, hull]
    :return: 1d np.array representing perimeter length or area of polygon on each frame

    .. note::
       Modified from `Jérôme Richard <https://stackoverflow.com/questions/74812556/computing-quick-convex-hull-using-numba/74817179#74817179>`_

    .. image:: _static/img/jitted_hull.png
       :width: 400
       :align: center



    :example:
    >>> points = np.random.randint(1, 50, size=(50, 5, 2)).astype(np.float32)
    >>> results = jitted_hull(points, target='area')
    """

    def perimeter(xy):
        perimeter = np.linalg.norm(xy[0] - xy[-1])
        for i in prange(xy.shape[0] - 1):
            p = np.linalg.norm(xy[i] - xy[i + 1])
            perimeter += p
        return perimeter

    def area(x, y):
        return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))

    results = np.full((points.shape[0]), np.nan)
    for i in range(points.shape[0]):
        S = points[i, :, :]
        a, b = np.argmin(S[:, 0]), np.argmax(S[:, 0])
        max_index = np.argmax(S[:, 0])
        idx = (
            process(S, np.arange(S.shape[0]), a, max_index)[:-1]
            + process(S, np.arange(S.shape[0]), max_index, a)[:-1]
        )
        x, y = np.full((len(idx)), np.nan), np.full((len(idx)), np.nan)
        for j in prange(len(idx)):
            x[j], y[j] = S[idx[j], 0], S[idx[j], 1]
        x0, y0 = np.mean(x), np.mean(y)
        r = np.sqrt((x - x0) ** 2 + (y - y0) ** 2)
        angles = np.where(
            (y - y0) > 0, np.arccos((x - x0) / r), 2 * np.pi - np.arccos((x - x0) / r)
        )
        mask = np.argsort(angles)
        x_sorted, y_sorted = x[mask], y[mask]
        if target == "perimeter":
            xy = np.vstack((x_sorted, y_sorted)).T
            results[i] = perimeter(xy)
        if target == "area":
            results[i] = area(x_sorted, y_sorted)

    return results


@jit(nopython=True)
def jitted_centroid(points: np.ndarray) -> np.ndarray:
    """
    Compute the centroid of polygons.

    :param array points: 3d array FRAMESxBODY-PARTxCOORDINATE
    :param str target: Options [perimeter, area]
    :return 1d np.array

    :example:
    >>> points = np.random.randint(1, 50, size=(50, 5, 2)).astype(float)
    >>> results = jitted_centroid(points)
    """

    results = np.full((points.shape[0], 2), np.nan)
    for i in range(points.shape[0]):
        S = points[i, :, :]
        a, b = np.argmin(S[:, 0]), np.argmax(S[:, 0])
        max_index = np.argmax(S[:, 0])
        idx = (
            process(S, np.arange(S.shape[0]), a, max_index)[:-1]
            + process(S, np.arange(S.shape[0]), max_index, a)[:-1]
        )
        perimeter_points = np.full((len(idx), 2), np.nan)
        for j in prange(len(idx)):
            perimeter_points[j] = points[i][j]
        results[i][0] = np.int(np.mean(perimeter_points[:, 0].flatten()))
        results[i][1] = np.int(np.mean(perimeter_points[:, 1].flatten()))
    return results


# points = np.random.randint(1, 5, size=(1, 10, 2)).astype(np.float32)
# points[0][1] = np.nan
# results = jitted_hull(points, target='area')
# print(results)


# points = np.random.randint(1, 10, size=(50, 5, 2)).astype(np.float32)
# results = jitted_centroid(points)
#
#
# results = np.full((points.shape[0]), np.nan)



import json
import math
import os
import random
import statistics
import sys

import dash
import dash_color_picker as dcp
import dash_colorscales as dcs
import dash_core_components as dcc
import dash_daq as daq
import dash_html_components as html
import h5py
import numpy as np
import pandas as pd
import plotly.graph_objects as go
from dash.dependencies import Input, Output, State

#
# FILE_PATH = sys.argv[1]
# GROUP_PATH = sys.argv[2]
FILE_PATH = "/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/logs/SimBA_dash_storage_20230202105812.h5"
GROUP_PATH = None

external_stylesheets = [
    "dash_simba_base.css",
    "//use.fontawesome.com/releases/v5.0.7/css/all.css",
]
app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
FONTS = [
    "Verdana",
    "Helvetica",
    "Calibri",
    "Arial",
    "Arial Narrow",
    "Candara",
    "Geneva",
    "Courier New",
    "Times New Roman",
]

GRAPH_CONFIG = {
    "toImageButtonOptions": {
        "format": "svg",
        "filename": "custom_image",
        "height": 500,
        "width": 700,
        "scale": 1,
    }
}


class DashApp(object):
    def __init__(self):
        self.groups_exist = False
        self.h5_file_path, self.group_path = FILE_PATH, GROUP_PATH
        self.graph_config = GRAPH_CONFIG
        self.video_info_df = pd.read_hdf(self.h5_file_path, key="Video_info")
        self.video_names = list(self.video_info_df.index)
        self.clf_names = list(pd.read_hdf(self.h5_file_path, key="Classifier_names"))
        if os.path.isfile(self.group_path):
            self.get_groups()
        self.fps = dict(zip(self.video_names, self.video_info_df["fps"]))
        self.run()

    def get_groups(self):
        self.groups_exist = True
        self.groups = {}
        self.groups_lookup = {}
        df = pd.read_csv(self.group_path).dropna()
        for group_id in df["GROUP"].unique():
            self.groups[group_id] = df[df["GROUP"] == group_id].tolist()

    def set_display(self):
        if self.groups_exist:
            return {"display": "block"}
        else:
            return {"display": "none"}

    def run(self):
        app.layout = html.Div(
            children=[
                html.Div(
                    id="header",
                    className="row",
                    children=[
                        html.H1(
                            children="SimBA Interactive Data Visualization Dashboard",
                            className="nine columns",
                        ),
                        html.Img(
                            src=app.get_asset_url("TheGoldenLab.PNG"),
                            className="three columns",
                            style={
                                "height": "15%",
                                "width": "15%",
                                "float": "right",
                                "position": "relative",
                            },
                        ),
                    ],
                ),
                html.Div(
                    id="simba-control-tabs",
                    className="three columns",
                    children=[
                        dcc.Tabs(
                            id="simba-tabs",
                            value="all_data",
                            children=[
                                dcc.Tab(
                                    label="Data",
                                    value="all_data",
                                    children=[
                                        html.Div(
                                            id="plot_data",
                                            children=[
                                                html.H6("Select Behavior to Plot"),
                                                html.Label("Behaviors:"),
                                                dcc.Dropdown(
                                                    # className='eight columns',
                                                    id="behaviors",
                                                    options=[
                                                        {"label": b, "value": b}
                                                        for b in self.clf_names
                                                    ],
                                                    multi=False,
                                                    value=self.clf_names[0],
                                                    clearable=False,
                                                ),
                                                html.Label("Category:"),
                                                dcc.Dropdown(
                                                    id="category_selector",
                                                    # className='eight columns',
                                                    clearable=False,
                                                ),
                                                html.Label("Feature:"),
                                                dcc.Dropdown(
                                                    id="feature_selector",
                                                    clearable=False,
                                                ),
                                                html.Div(
                                                    id="multi_group_data",
                                                    children=[
                                                        html.Hr(),
                                                        html.H6("Plotting Group Means"),
                                                        html.Label("Select Group(s):"),
                                                        dcc.Dropdown(
                                                            id="multi_group_selector",
                                                            # className='eight columns',
                                                            options=[
                                                                {
                                                                    "label": key,
                                                                    "value": key,
                                                                }
                                                                for key in self.groups.keys()
                                                            ],
                                                            multi=True,
                                                            clearable=False,
                                                            value=self.groups,
                                                        ),
                                                        dcc.Checklist(
                                                            id="group_mean",
                                                            options=[
                                                                {
                                                                    "label": "Show Total Mean",
                                                                    "value": "showGroupMean",
                                                                    "disabled": False,
                                                                }
                                                            ],
                                                            value=["showGroupMean"],
                                                            labelStyle={
                                                                "display": "inline-block"
                                                            },
                                                        ),
                                                    ],
                                                    style=self.set_display(),
                                                ),
                                                html.Hr(),
                                                html.H6("Plotting Individual Groups"),
                                                html.Div(
                                                    id="show_group_opt",
                                                    children=[
                                                        html.Label(
                                                            "Select Single Group to Plot:",
                                                            style=self.set_display(),
                                                        ),
                                                        dcc.Dropdown(
                                                            id="group_selector",
                                                            options=[
                                                                {
                                                                    "label": key,
                                                                    "value": key,
                                                                }
                                                                for key in groups_dict.keys()
                                                            ],
                                                            multi=False,
                                                            clearable=False,
                                                        ),
                                                    ],
                                                    style=self.set_display(),
                                                ),
                                                html.Label("Select Video(s):"),
                                                dcc.Dropdown(
                                                    id="video_numbers",
                                                    multi=True,
                                                    clearable=False,
                                                ),
                                                dcc.Checklist(
                                                    id="video_mean",
                                                    options=[
                                                        {
                                                            "label": "Show Group Mean",
                                                            "value": "showVideoMean",
                                                            "disabled": False,
                                                        }
                                                    ],
                                                    value=["showVideoMean"],
                                                ),
                                            ],
                                        )
                                    ],
                                ),
                                dcc.Tab(
                                    label="Graph Settings",
                                    value="graph_settings",
                                    children=[
                                        html.Div(
                                            className="twelve columns",
                                            id="graph_settings",
                                            children=[
                                                html.Div(
                                                    id="color_properties",
                                                    className="twelve columns",
                                                    children=[
                                                        html.H6("Color Properties"),
                                                        html.Div(
                                                            id="color_pickers",
                                                            children=create_color_pickers(),
                                                        ),
                                                        html.Label(
                                                            "Colorscale Gradient for Individual Videos"
                                                        ),
                                                        dcs.DashColorscales(
                                                            id="ind_group_colors",
                                                            colorscale=[
                                                                "#fbb4ae",
                                                                "#b3cde3",
                                                                "#ccebc5",
                                                                "#decbe4",
                                                                "#fed9a6",
                                                                "#ffffcc",
                                                            ],
                                                        ),
                                                        html.Button(
                                                            "Update Colors",
                                                            id="update_graph",
                                                            n_clicks=0,
                                                            type="submit",
                                                        ),
                                                        html.Hr(),
                                                    ],
                                                ),
                                                html.Div(
                                                    className="twelve columns",
                                                    id="probability-graph-properties",
                                                    children=[
                                                        html.H6(
                                                            "Probability Graph Properties"
                                                        ),
                                                        html.Div(
                                                            id="multi_prob_prop",
                                                            children=[
                                                                html.Label(
                                                                    "Set Group Viewing Frames:"
                                                                ),
                                                                dcc.Input(
                                                                    id="multi_start_x",
                                                                    type="number",
                                                                    className="four columns",
                                                                ),
                                                                dcc.Input(
                                                                    id="multi_end_x",
                                                                    type="number",
                                                                    className="four columns",
                                                                ),
                                                                html.Button(
                                                                    id="submit_multi_range",
                                                                    className="three columns",
                                                                    n_clicks=1,
                                                                    children="Submit",
                                                                    type="submit",
                                                                ),
                                                                html.Label(
                                                                    "Multi Group (Seconds)"
                                                                ),
                                                                dcc.Input(
                                                                    id="multi_start_sec",
                                                                    type="number",
                                                                    className="four columns",
                                                                    disabled=True,
                                                                ),
                                                                dcc.Input(
                                                                    id="multi_end_sec",
                                                                    type="number",
                                                                    className="four columns",
                                                                    disabled=True,
                                                                ),
                                                                html.Button(
                                                                    id="reset_multi_range",
                                                                    className="twelve columns",
                                                                    n_clicks=1,
                                                                    children="Reset Axes",
                                                                    type="reset",
                                                                ),
                                                            ],
                                                            style=self.set_display(),
                                                        ),
                                                        html.Label(
                                                            "Set Video Viewing Frames:",
                                                            className="twelve columns",
                                                        ),
                                                        dcc.Input(
                                                            id="ind_start_x",
                                                            type="number",
                                                            className="four columns",
                                                        ),
                                                        dcc.Input(
                                                            id="ind_end_x",
                                                            type="number",
                                                            className="four columns",
                                                        ),
                                                        html.Button(
                                                            id="submit_ind_range",
                                                            className="three columns",
                                                            n_clicks=1,
                                                            children="Submit",
                                                            type="submit",
                                                        ),
                                                        html.Label(
                                                            "Individual Group (Seconds)",
                                                            className="twelve columns",
                                                        ),
                                                        dcc.Input(
                                                            id="ind_start_sec",
                                                            type="number",
                                                            className="four columns",
                                                            disabled=True,
                                                        ),
                                                        dcc.Input(
                                                            id="ind_end_sec",
                                                            type="number",
                                                            className="four columns",
                                                            disabled=True,
                                                        ),
                                                        html.Button(
                                                            id="reset_ind_range",
                                                            className="twelve columns",
                                                            n_clicks=1,
                                                            children="Reset Axes",
                                                            type="reset",
                                                        ),
                                                    ],
                                                    style={"display": "block"},
                                                ),
                                                html.Div(
                                                    className="twelve columns",
                                                    id="bar-graph-properties",
                                                    children=[
                                                        html.H6("Bar Graph Properties"),
                                                        html.Label("Error Type:"),
                                                        dcc.RadioItems(
                                                            id="error_type",
                                                            options=[
                                                                {
                                                                    "label": "Above",
                                                                    "value": "above",
                                                                },
                                                                {
                                                                    "label": "Below",
                                                                    "value": "below",
                                                                },
                                                                {
                                                                    "label": "Both",
                                                                    "value": "both",
                                                                },
                                                            ],
                                                            value="both",
                                                        ),
                                                    ],
                                                ),
                                                html.Div(
                                                    className="twelve columns",
                                                    id="additional-properties",
                                                    children=[
                                                        html.Hr(),
                                                        html.H6(
                                                            "Additional Properties"
                                                        ),
                                                        dcc.Checklist(
                                                            id="graph_seconds",
                                                            options=[
                                                                {
                                                                    "label": "Graph X Axis in Seconds",
                                                                    "value": "show_seconds",
                                                                }
                                                            ],
                                                            value=[],
                                                            labelStyle={
                                                                "display": "inline-block"
                                                            },
                                                        ),
                                                        dcc.Checklist(
                                                            id="show_grid",
                                                            options=[
                                                                {
                                                                    "label": "Show Grid lines",
                                                                    "value": "showGridLines",
                                                                }
                                                            ],
                                                            value=["showGridLines"],
                                                            labelStyle={
                                                                "display": "inline-block"
                                                            },
                                                        ),
                                                        dcc.Checklist(
                                                            id="show_bg",
                                                            options=[
                                                                {
                                                                    "label": "Show Background",
                                                                    "value": "showBg",
                                                                }
                                                            ],
                                                            value=["showBg"],
                                                            labelStyle={
                                                                "display": "inline-block"
                                                            },
                                                        ),
                                                        html.Div(
                                                            id="show_mean_graph_title",
                                                            children=[
                                                                html.Label(
                                                                    "Group Means Title:"
                                                                ),
                                                                dcc.Input(
                                                                    id="mean_graph_title",
                                                                    className="seven columns",
                                                                ),
                                                                html.Button(
                                                                    children=[
                                                                        html.I(
                                                                            n_clicks=0,
                                                                            className="fa fa-chart-area",
                                                                            style={
                                                                                "font-size": "24px"
                                                                            },
                                                                        ),
                                                                        html.P(
                                                                            "Set",
                                                                            style={
                                                                                "padding-left": "5px"
                                                                            },
                                                                        ),
                                                                    ],
                                                                    id="set_multi_title",
                                                                    n_clicks=0,
                                                                    className="four columns",
                                                                ),
                                                            ],
                                                            style=self.set_display(),
                                                        ),
                                                        html.Label(
                                                            "Individual Videos Title:"
                                                        ),
                                                        dcc.Input(
                                                            id="ind_graph_title",
                                                            className="seven columns",
                                                        ),
                                                        html.Button(
                                                            children=[
                                                                html.I(
                                                                    n_clicks=0,
                                                                    className="fa fa-chart-area",
                                                                    style={
                                                                        "font-size": "24px"
                                                                    },
                                                                ),
                                                                html.P(
                                                                    "Set",
                                                                    style={
                                                                        "padding-left": "5px"
                                                                    },
                                                                ),
                                                            ],
                                                            id="set_ind_title",
                                                            n_clicks=0,
                                                            className="four columns",
                                                        ),
                                                        html.Label("Choose Font:"),
                                                        dcc.Dropdown(
                                                            id="fonts",
                                                            options=[
                                                                {"label": f, "value": f}
                                                                for f in FONTS
                                                            ],
                                                            value=FONTS[0],
                                                            multi=False,
                                                            clearable=False,
                                                        ),
                                                        html.Label("Font Size:"),
                                                        daq.NumericInput(
                                                            id="font_size",
                                                            value=12,
                                                            min=8,
                                                            max=48,
                                                        ),
                                                    ],
                                                ),
                                            ],
                                        )
                                    ],
                                ),
                                dcc.Tab(
                                    label="Download Settings",
                                    value="download_settings",
                                    children=[
                                        html.Div(
                                            id="download-settings",
                                            children=[
                                                html.Div(
                                                    id="csv_export",
                                                    className="twelve columns",
                                                    children=[
                                                        html.H6("CSV Export"),
                                                        html.Label(
                                                            "Enter csv file name:"
                                                        ),
                                                        dcc.Input(id="csv_file_name"),
                                                        html.P(children=[".csv"]),
                                                        html.Button(
                                                            children=[
                                                                html.I(
                                                                    n_clicks=0,
                                                                    className="fa fa-download",
                                                                    style={
                                                                        "font-size": "24px"
                                                                    },
                                                                ),
                                                                html.P(
                                                                    "Group Means",
                                                                    style={
                                                                        "padding-left": "5px"
                                                                    },
                                                                ),
                                                            ],
                                                            id="download_group_data",
                                                            n_clicks=0,
                                                            style=self.set_display(),
                                                            className="five columns",
                                                        ),
                                                        html.Button(
                                                            children=[
                                                                html.I(
                                                                    n_clicks=0,
                                                                    className="fa fa-download",
                                                                    style={
                                                                        "font-size": "24px"
                                                                    },
                                                                ),
                                                                html.P(
                                                                    "Videos",
                                                                    style={
                                                                        "padding-left": "5px"
                                                                    },
                                                                ),
                                                            ],
                                                            id="download_data",
                                                            n_clicks=0,
                                                            className="five columns",
                                                        ),
                                                        html.Div(
                                                            id="comments",
                                                            className="twelve columns",
                                                        ),
                                                        html.Hr(
                                                            className="twelve columns"
                                                        ),
                                                    ],
                                                ),
                                                html.Div(
                                                    id="img_export",
                                                    className="twelve columns",
                                                    children=[
                                                        html.H6(
                                                            "Image Export",
                                                            className="twelve columns",
                                                        ),
                                                        html.Div(
                                                            id="img-dim",
                                                            className="twelve columns",
                                                            children=[
                                                                html.Label(
                                                                    "Enter image dimensions (px):",
                                                                    className="twelve columns",
                                                                ),
                                                                daq.NumericInput(
                                                                    id="img_height",
                                                                    className="three columns",
                                                                    value=500,
                                                                    min=0,
                                                                    max=1000,
                                                                ),
                                                                html.P(
                                                                    "x",
                                                                    className="one column",
                                                                ),
                                                                daq.NumericInput(
                                                                    id="img_width",
                                                                    className="three columns",
                                                                    value=700,
                                                                    min=0,
                                                                    max=1000,
                                                                ),
                                                            ],
                                                        ),
                                                        html.Label(
                                                            "Enter image name:",
                                                            className="twelve columns",
                                                        ),
                                                        dcc.Input(id="img_file_name"),
                                                        html.P(
                                                            id="img_type",
                                                            children=[".svg"],
                                                        ),
                                                        html.Label("File extension:"),
                                                        dcc.Dropdown(
                                                            id="img_ext",
                                                            options=[
                                                                {
                                                                    "label": "PNG",
                                                                    "value": "png",
                                                                },
                                                                {
                                                                    "label": "JPEG",
                                                                    "value": "jpeg",
                                                                },
                                                                {
                                                                    "label": "SVG",
                                                                    "value": "svg",
                                                                },
                                                            ],
                                                            value="svg",
                                                        ),
                                                        html.Button(
                                                            children=[
                                                                html.I(
                                                                    n_clicks=0,
                                                                    className="fa fa-save",
                                                                    style={
                                                                        "font-size": "24px"
                                                                    },
                                                                ),
                                                                html.P(
                                                                    "Save Download Settings",
                                                                    style={
                                                                        "padding-left": "5px"
                                                                    },
                                                                ),
                                                            ],
                                                            id="submit_download",
                                                            n_clicks=0,
                                                            type="submit",
                                                        ),
                                                    ],
                                                ),
                                            ],
                                        )
                                    ],
                                ),
                            ],
                        )
                    ],
                ),
                html.Div(
                    id="graphs",
                    className="eight columns",
                    children=[
                        html.Div(
                            id="multi_group_container",
                            # style={'display': 'none'},
                            children=[
                                dcc.Graph(
                                    id="plot_multi_groups",
                                    style={"display": "none"},
                                    config=self.config,
                                )
                            ],
                        ),
                        dcc.Graph(
                            id="plot_ind_group",
                            config=self.config,
                            style={"display": "block"},
                        ),
                    ],
                ),
                html.Div(
                    id="credits",
                    className="footer",
                    children=[
                        html.Div(
                            className="six columns",
                            children=[
                                html.P("Created by "),
                                html.A(
                                    "Sophia Hwang",
                                    href="https://github.com/sophihwang26",
                                    target="_blank",
                                ),
                                html.P(" and "),
                                html.A(
                                    "Aasiya Islam",
                                    href="https://github.com/aasiya-islam",
                                    target="_blank",
                                ),
                            ],
                        ),
                        html.Div(
                            className="six columns",
                            children=[
                                html.P(
                                    "Built with Dash",
                                    style={
                                        "height": "15%",
                                        "width": "15%",
                                        "float": "right",
                                        "position": "relative",
                                    },
                                )
                            ],
                        ),
                    ],
                ),
            ]
        )


@app.callback(Output("img_type", "children"), [Input("img_ext", "value")])
def update_ext_text(extension):
    return [".{}".format(extension)]


if __name__ == "__main__":
    app.run_server(debug=False)



import pytest
import os
from simba.data_processors.agg_clf_calculator import AggregateClfCalculator
from simba.data_processors.fsttc_calculator import FSTTCCalculator
from simba.data_processors.kleinberg_calculator import KleinbergCalculator
from simba.data_processors.movement_calculator import MovementCalculator
from simba.data_processors.timebins_clf_calculator import TimeBinsClfCalculator
from simba.data_processors.timebins_movement_calculator import TimeBinsMovementCalculator




@pytest.mark.parametrize("config_path, data_measures, video_meta_data, classifiers", [('tests/data/test_projects/two_c57/project_folder/project_config.ini', ['Bout count'], [], ['Attack']),
                                                                                      ('tests/data/test_projects/two_c57/project_folder/project_config.ini', ['Total event duration (s)'], [], ['Attack'])])
def test_create_clf_log_use_case(config_path, data_measures, video_meta_data, classifiers):

    clf_log_creator = AggregateClfCalculator(config_path=config_path,
                                             data_measures=data_measures, 
                                             classifiers=classifiers,
                                             video_meta_data=video_meta_data)
    clf_log_creator.run()
    clf_log_creator.save()
    assert len(clf_log_creator.results_df) == len(data_measures)
    assert os.path.isfile(clf_log_creator.save_path)
    os.remove(clf_log_creator.save_path)
    

@pytest.mark.parametrize("config_path, time_window, behavior_lst, create_graph", [('tests/data/test_projects/two_c57/project_folder/project_config.ini', 2000, ['Attack', 'Sniffing'], False)])
def test_fsttc_calculator_use_case(config_path, time_window, behavior_lst, create_graph):
    fsttc_calculator = FSTTCCalculator(config_path=config_path, time_window=time_window, behavior_lst=behavior_lst, create_graphs=create_graph)
    fsttc_calculator.run()
    assert len(fsttc_calculator.out_df) == len(behavior_lst)
    if create_graph:
        assert os.path.isfile(fsttc_calculator.save_plot_path)
    os.remove(fsttc_calculator.file_save_path)

@pytest.mark.parametrize("config_path, classifier_names, sigma, gamma, hierarchy, hierarchical_search", [('tests/data/test_projects/two_c57/project_folder/project_config.ini', ['Attack'], 1.1, 0.3, 5, False)])
def test_kleinberg_calculator_use_case(config_path, classifier_names, gamma, sigma, hierarchy, hierarchical_search):
    calculator = KleinbergCalculator(config_path=config_path,
                                     classifier_names=classifier_names,
                                     sigma=sigma,
                                     gamma=gamma,
                                     hierarchy=hierarchy,
                                     hierarchical_search=hierarchical_search)
    calculator.run()


@pytest.mark.parametrize("config_path, body_parts, threshold", [('tests/data/test_projects/two_c57/project_folder/project_config.ini', ['simon CENTER OF GRAVITY'], 0.0),
                                                                ('tests/data/test_projects/two_c57/project_folder/project_config.ini', ['Nose_1'], 0.0)])
def test_movement_calculator_use_case(config_path, body_parts, threshold):
    calculator = MovementCalculator(config_path=config_path,
                                    body_parts=body_parts,
                                    threshold=threshold)
    calculator.run()
    calculator.save()
    assert os.path.isfile(calculator.save_path)
    os.remove(calculator.save_path)


@pytest.mark.parametrize("config_path, bin_length, measurements, classifiers", [('tests/data/test_projects/two_c57/project_folder/project_config.ini', 2, ['Event count'], ['Attack'])])
def test_time_bins_clf_calculator_use_case(config_path, bin_length, measurements, classifiers):
    calculator = TimeBinsClfCalculator(config_path=config_path,
                                       bin_length=bin_length,
                                       measurements=measurements,
                                       classifiers=classifiers)
    calculator.run()
    assert os.path.isfile(calculator.save_path)
    os.remove(calculator.save_path)


@pytest.mark.parametrize("config_path, bin_length, plots, body_parts", [('tests/data/test_projects/two_c57/project_folder/project_config.ini', 10, True, ['Nose_1', 'Nose_2'])])
def test_time_bins_movement_calculator_use_case(config_path, bin_length, plots, body_parts):
    calculator = TimeBinsMovementCalculator(config_path=config_path,
                                            bin_length=bin_length,
                                            plots=plots,
                                            body_parts=body_parts)
    calculator.run()
    assert os.path.isfile(calculator.save_path); os.remove(calculator.save_path)
    if plots:
        assert os.path.isfile(calculator.plot_save_path); os.remove(calculator.plot_save_path)



__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
import platform
import shutil
from typing import Dict, List

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_that_column_exist
from simba.utils.enums import Formats
from simba.utils.errors import ColumnNotFoundError, NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    read_df)


class TresholdPlotCreatorMultiprocess(ConfigReader, PlottingMixin):
    """
    Class for line chart visualizations displaying the classification probabilities of a single classifier.
    Uses multiprocessing.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str clf_name: Name of the classifier to create visualizations for
    :param bool frame_setting: When True, SimBA creates indidvidual frames in png format
    :param bool video_setting: When True, SimBA creates compressed video in mp4 format
    :param bool last_image: When True, creates image .png representing last frame of the video.
    :param dict style_attr: User-defined style attributes of the visualization (line size, color etc).
    :param List[str] files_found: Files to create threshold plots for.
    :param int cores: Number of cores to use.

    .. note::
       `Visualization tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. image:: _static/img/prob_plot.png
       :width: 300
       :align: center

    Examples
    ----------
    >>> plot_creator = TresholdPlotCreatorMultiprocess(config_path='/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini', frame_setting=True, video_setting=True, clf_name='Attack', style_attr={'width': 640, 'height': 480, 'font size': 10, 'line width': 6, 'color': 'magneta', 'circle size': 20}, cores=5)
    >>> plot_creator.run()
    """

    def __init__(
        self,
        config_path: str,
        clf_name: str,
        frame_setting: bool,
        video_setting: bool,
        last_frame: bool,
        cores: int,
        style_attr: Dict[str, int],
        files_found: List[str],
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)
        (
            self.frame_setting,
            self.video_setting,
            self.cores,
            self.style_attr,
            self.last_frame,
        ) = (frame_setting, video_setting, cores, style_attr, last_frame)
        if (
            (not self.frame_setting)
            and (not self.video_setting)
            and (not self.last_frame)
        ):
            raise NoSpecifiedOutputError(
                "SIMBA ERROR: Please choose to create either probability videos, frames, and/or last frame.",
                show_window=True,
            )
        self.clf_name, self.files_found = clf_name, files_found
        self.probability_col = "Probability_" + self.clf_name
        self.fontsize = self.style_attr["font size"]
        self.out_width, self.out_height = (
            self.style_attr["width"],
            self.style_attr["height"],
        )
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        if not os.path.exists(self.probability_plot_dir):
            os.makedirs(self.probability_plot_dir)
        print(f"Processing {str(len(self.files_found))} video(s)...")

    def run(self):
        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            video_info, self.px_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            data_df = read_df(file_path, self.file_type)
            if self.probability_col not in data_df.columns:
                raise ColumnNotFoundError(
                    column_name=self.probability_col, file_name=file_path
                )
            check_that_column_exist(
                df=data_df, column_name=self.clf_name, file_name=self.video_name
            )
            self.save_frame_folder_dir = os.path.join(
                self.probability_plot_dir, self.video_name + "_" + self.clf_name
            )
            self.video_folder = os.path.join(
                self.probability_plot_dir, self.video_name + "_" + self.clf_name
            )
            self.temp_folder = os.path.join(
                self.probability_plot_dir, self.video_name + "_" + self.clf_name, "temp"
            )
            if self.frame_setting:
                if os.path.exists(self.save_frame_folder_dir):
                    shutil.rmtree(self.save_frame_folder_dir)
                if not os.path.exists(self.save_frame_folder_dir):
                    os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                if os.path.exists(self.temp_folder):
                    shutil.rmtree(self.temp_folder)
                    shutil.rmtree(self.video_folder)
                os.makedirs(self.temp_folder)
                self.save_video_path = os.path.join(
                    self.probability_plot_dir,
                    "{}_{}.mp4".format(self.video_name, self.clf_name),
                )

            probability_lst = list(data_df[self.probability_col])

            if self.last_frame:
                _ = self.make_probability_plot(
                    data=pd.Series(probability_lst),
                    style_attr=self.style_attr,
                    clf_name=self.clf_name,
                    fps=self.fps,
                    save_path=os.path.join(
                        self.probability_plot_dir,
                        self.video_name
                        + "_{}_{}.png".format(self.clf_name, "final_image"),
                    ),
                )

            if self.video_setting or self.frame_setting:
                if self.style_attr["y_max"] == "auto":
                    highest_p = data_df[self.probability_col].max()
                else:
                    highest_p = float(self.style_attr["y_max"])
                data_split = np.array_split(list(data_df.index), self.cores)
                frm_per_core = len(data_split[0])
                for group_cnt, rng in enumerate(data_split):
                    data_split[group_cnt] = np.insert(rng, 0, group_cnt)

                print(
                    "Creating probability images, multiprocessing (determined chunksize: {}, cores: {})...".format(
                        str(self.multiprocess_chunksize), str(self.cores)
                    )
                )
                with multiprocessing.Pool(
                    self.cores, maxtasksperchild=self.maxtasksperchild
                ) as pool:
                    constants = functools.partial(
                        self.probability_plot_mp,
                        clf_name=self.clf_name,
                        probability_lst=probability_lst,
                        highest_p=highest_p,
                        video_setting=self.video_setting,
                        frame_setting=self.frame_setting,
                        fps=self.fps,
                        video_dir=self.temp_folder,
                        frame_dir=self.save_frame_folder_dir,
                        style_attr=self.style_attr,
                        video_name=self.video_name,
                    )
                    for cnt, result in enumerate(
                        pool.imap(
                            constants, data_split, chunksize=self.multiprocess_chunksize
                        )
                    ):
                        print(
                            "Image {}/{}, Video {}/{}...".format(
                                str(int(frm_per_core * (result + 1))),
                                str(len(data_df)),
                                str(file_cnt + 1),
                                str(len(self.files_found)),
                            )
                        )

                pool.join()
                pool.terminate()
                if self.video_setting:
                    print("Joining {} multiprocessed video...".format(self.video_name))
                    concatenate_videos_in_folder(
                        in_folder=self.temp_folder, save_path=self.save_video_path
                    )

                video_timer.stop_timer()
                print(
                    "Probability video {} complete (elapsed time: {}s) ...".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )

        self.timer.stop_timer()
        stdout_success(
            msg=f"Probability visualizations for {str(len(self.files_found))} videos created in project_folder/frames/output/gantt_plots directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = TresholdPlotCreatorMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals/project_folder/project_config.ini',
#                                         frame_setting=False,
#                                         video_setting=True,
#                                         last_frame=False,
#                                         clf_name='Attack',
#                                         cores=5,
#                                         files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals/project_folder/csv/machine_results/Together_1.csv'],
#                                         style_attr={'width': 640, 'height': 480, 'font size': 10, 'line width': 6, 'color': 'blue', 'circle size': 20, 'y_max': 'auto'})
# #test = TresholdPlotCreatorSingleProcess(config_path='/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini', frame_setting=False, video_setting=True, clf_name='Attack')
# test.create_plots()


# test = TresholdPlotCreatorMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                         frame_setting=False,
#                                         video_setting=True,
#                                         last_frame=True,
#                                         clf_name='Attack',
#                                         cores=5,
#                                         files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                         style_attr={'width': 640, 'height': 480, 'font size': 10, 'line width': 3, 'color': 'blue', 'circle size': 20, 'y_max': 'auto'})
# test.create_plots()



__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log, check_int,
    check_that_column_exist)
from simba.utils.data import detect_bouts
from simba.utils.enums import Formats, TagNames, TextOptions
from simba.utils.errors import NoFilesFoundError, NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder,
                                    find_core_cnt, get_fn_ext,
                                    get_video_meta_data, read_df, remove_files)
from simba.utils.warnings import NoDataFoundWarning

SPACE_SCALE = 60
RADIUS_SCALE = 12
RESOLUTION_SCALE = 1500
FONT_SCALE = 1.5


def val_clip_createror_mp(
    data: np.ndarray,
    video_path: str,
    scalers: dict,
    bout_total_cnt: int,
    fps: float,
    video_meta_data: dict,
    clf_name: str,
    p_data: pd.Series,
    clf_data: pd.Series,
    highlight_clr: tuple,
    text_clr: tuple,
):
    def __insert_inter_frms(
        bg_color: Tuple[int, int, int] = (49, 32, 189),
        fg_color: Tuple[int, int, int] = (0, 0, 0),
    ):
        """
        Helper to create N blank frames separating the classified event bouts with defined BGR colors.
        """
        for i in range(int(fps)):
            inter_frm = np.full(
                (int(video_meta_data["height"]), int(video_meta_data["width"]), 3),
                bg_color,
            ).astype(np.uint8)
            cv2.putText(
                inter_frm,
                f"Bout #{bount_cnt}",
                (
                    10,
                    (video_meta_data["height"] - video_meta_data["height"])
                    + scalers["space"],
                ),
                TextOptions.FONT.value,
                scalers["font"],
                fg_color,
                TextOptions.TEXT_THICKNESS.value,
            )
            writer.write(inter_frm)

    SPACER = 2
    cap = cv2.VideoCapture(video_path)
    fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
    start_frm, end_frame, save_path, c_frm, bount_cnt = (
        int(data[1]),
        int(data[2]),
        data[3],
        int(data[1]),
        int(data[0]),
    )
    bout_frm_cnt = end_frame - start_frm
    writer = cv2.VideoWriter(
        save_path,
        fourcc,
        fps,
        (int(video_meta_data["width"]), int(video_meta_data["height"])),
    )
    __insert_inter_frms()
    frm_cnt = 0
    while c_frm < end_frame:
        p, clf_val = round(float(p_data.loc[c_frm]), 3), int(clf_data.loc[c_frm])
        cap.set(1, c_frm)
        ret, img = cap.read()
        cv2.putText(
            img,
            f"{clf_name} event # {bount_cnt}",
            (
                TextOptions.BORDER_BUFFER_Y.value,
                (video_meta_data["height"] - video_meta_data["height"])
                + scalers["space"] * SPACER,
            ),
            TextOptions.FONT.value,
            scalers["font"],
            text_clr,
            TextOptions.TEXT_THICKNESS.value + 1,
        )
        SPACER += 1
        cv2.putText(
            img,
            f"Total frames of event: {end_frame-start_frm+1}",
            (
                TextOptions.BORDER_BUFFER_Y.value,
                (video_meta_data["height"] - video_meta_data["height"])
                + scalers["space"] * SPACER,
            ),
            TextOptions.FONT.value,
            scalers["font"],
            text_clr,
            TextOptions.TEXT_THICKNESS.value + 1,
        )
        SPACER += 1
        cv2.putText(
            img,
            f"Frames of event {start_frm} to {end_frame}",
            (
                TextOptions.BORDER_BUFFER_Y.value,
                (video_meta_data["height"] - video_meta_data["height"])
                + scalers["space"] * SPACER,
            ),
            TextOptions.FONT.value,
            scalers["font"],
            text_clr,
            TextOptions.TEXT_THICKNESS.value + 1,
        )
        SPACER += 1
        cv2.putText(
            img,
            f"Frame number: {c_frm}",
            (
                TextOptions.BORDER_BUFFER_Y.value,
                (video_meta_data["height"] - video_meta_data["height"])
                + scalers["space"] * SPACER,
            ),
            TextOptions.FONT.value,
            scalers["font"],
            text_clr,
            TextOptions.TEXT_THICKNESS.value + 1,
        )
        SPACER += 1
        if (highlight_clr != None) and (clf_val == 1):
            cv2.putText(
                img,
                f"Frame {clf_name} probability: {p}",
                (
                    TextOptions.BORDER_BUFFER_Y.value,
                    (video_meta_data["height"] - video_meta_data["height"])
                    + scalers["space"] * SPACER,
                ),
                TextOptions.FONT.value,
                scalers["font"],
                highlight_clr,
                TextOptions.TEXT_THICKNESS.value + 1,
            )
        else:
            cv2.putText(
                img,
                f"Frame {clf_name} probability: {p}",
                (
                    TextOptions.BORDER_BUFFER_Y.value,
                    (video_meta_data["height"] - video_meta_data["height"])
                    + scalers["space"] * SPACER,
                ),
                TextOptions.FONT.value,
                scalers["font"],
                text_clr,
                TextOptions.TEXT_THICKNESS.value + 1,
            )
        writer.write(img)
        c_frm += 1
        SPACER = 2
        frm_cnt += 1
        print(
            f'Multiprocessing frame {frm_cnt}/{bout_frm_cnt} (Bout: {bount_cnt+1}/{bout_total_cnt}, Video: {video_meta_data["video_name"]})'
        )
    writer.release()
    cap.release()


class ClassifierValidationClipsMultiprocess(ConfigReader):
    """
    Create video clips with overlaid classified events for detection of false positive event bouts using multiple cores for improved runtime.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter int window: Number of seconds before and after the event bout that should be included in the output video.
    :parameter str clf_name: Name of the classifier to create validation videos for.
    :parameter bool clips: If True, creates individual video file clips for each validation bout.
    :parameter List[Union[str, os.PathLike]] data_paths: List of files with classification results to create videos for.
    :parameter Tuple[int, int, int] text_clr: Color of text overlay in BGR.
    :parameter Optional[Tuple[int, int, int]] highlight_clr: Color of text when probability values are above threshold. If None, same as text_clr.
    :parameter float video_speed:  FPS rate in relation to original video. E.g., the same as original video if 1.0. If output should be half the speed relative to input, set to 0.5. Default: 1.0.
    :parameter bool concat_video:  If True, creates a single video including all events bouts for each video. Default: False.
    :parameter Optional[int] core_cnt: Integer dictating the numbers of cores to use. If -1, all available cores are used.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/classifier_validation.md#classifier-validation>`_.

    Examples
    ----------
    >>> _ = ClassifierValidationClipsMultiprocess(config_path='MyProjectConfigPath', window=5, clf_name='Attack', text_clr=(255,255,0), clips=False, concat_video=True).run()
    """

    def __init__(
        self,
        config_path: str,
        window: int,
        clf_name: str,
        clips: bool,
        data_paths: List[Union[str, os.PathLike]],
        text_clr: Tuple[int, int, int],
        concat_video: bool = False,
        video_speed: float = 1.0,
        highlight_clr: Optional[Tuple[int, int, int]] = None,
        core_cnt: Optional[int] = -1,
    ):
        super().__init__(config_path=config_path)
        log_event(
            logger_name=str(__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        if (not clips) and (not concat_video):
            raise NoSpecifiedOutputError(
                msg="Please select to create clips and/or a concatenated video",
                source=self.__class__.__name__,
            )

        check_int(name="Time window", value=window, min_value=0)
        check_int(
            name="CORE COUNT",
            value=core_cnt,
            min_value=-1,
            max_value=find_core_cnt()[0],
            raise_error=True,
        )
        if core_cnt == -1:
            core_cnt = find_core_cnt()[0]
        self.core_cnt = core_cnt
        self.window, self.clf_name = int(window), clf_name
        self.clips, self.concat_video, self.video_speed, self.highlight_clr = (
            clips,
            concat_video,
            video_speed,
            highlight_clr,
        )
        self.p_col = f"Probability_{self.clf_name}"
        self.text_clr, self.data_paths = text_clr, data_paths
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.font = TextOptions.FONT.value
        if not os.path.exists(self.clf_validation_dir):
            os.makedirs(self.clf_validation_dir)
        if len(self.data_paths) == 0:
            raise NoFilesFoundError(
                msg="No data found in the project_folder/csv/machine_results directory",
                source=self.__class__.__name__,
            )
        print(f"Processing {str(len(self.data_paths))} files...")

    def run(self):
        """
        Method to generate clips. Results are saved in the ``project_folder/frames/output/classifier_validation directory``
        directory of the SimBA project.
        """

        check_all_file_names_are_represented_in_video_log(
            video_info_df=self.video_info_df, data_paths=self.data_paths
        )

        for file_cnt, file_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            self.data_df = read_df(file_path, self.file_type)
            check_that_column_exist(
                df=self.data_df, column_name=self.p_col, file_name=file_path
            )
            _, file_name, _ = get_fn_ext(file_path)
            self.video_path = self.find_video_of_file(
                video_dir=self.video_dir, filename=file_name, raise_error=True
            )
            self.video_info = get_video_meta_data(video_path=self.video_path)
            self.fps = int(self.video_info["fps"])
            self.video_fps = int(self.fps * self.video_speed)
            if self.video_fps < 1:
                self.video_fps = 1
            self.max_dim = max(self.video_info["width"], self.video_info["height"])
            self.circle_scale = int(RADIUS_SCALE / (RESOLUTION_SCALE / self.max_dim))
            self.font_size = float(FONT_SCALE / (RESOLUTION_SCALE / self.max_dim))
            self.spacing_scale = int(SPACE_SCALE / (RESOLUTION_SCALE / self.max_dim))
            clf_bouts = detect_bouts(
                data_df=self.data_df, target_lst=[self.clf_name], fps=self.fps
            ).reset_index(drop=True)
            if len(clf_bouts) == 0:
                NoDataFoundWarning(
                    msg=f"Skipping video {file_name}: No classified behavior detected...",
                    source=self.__class__.__name__,
                )
                continue

            clip_data = []
            for i, (bout_cnt, bout) in enumerate(clf_bouts.iterrows()):
                self.bout_cnt = bout_cnt
                event_start_frm, event_end_frm = bout["Start_frame"], bout["End_frame"]
                start_window = int(
                    event_start_frm - (int(self.video_info["fps"]) * self.window)
                )
                end_window = int(
                    event_end_frm + (int(self.video_info["fps"]) * self.window)
                )
                if end_window > len(self.data_df):
                    end_window = len(self.data_df)
                if start_window < 0:
                    start_window = 0
                self.save_path = os.path.join(
                    self.clf_validation_dir,
                    self.clf_name + f"_{bout_cnt}_{file_name}.mp4",
                )
                clip_data.append([bout_cnt, start_window, end_window, self.save_path])
            clip_data = np.array(clip_data)
            print(
                f"Creating validation video, multiprocessing (chunksize: {self.multiprocess_chunksize}, cores: {self.core_cnt})..."
            )
            with multiprocessing.Pool(
                self.core_cnt, maxtasksperchild=self.maxtasksperchild
            ) as pool:
                constants = functools.partial(
                    val_clip_createror_mp,
                    video_path=self.video_path,
                    scalers={
                        "circle": self.circle_scale,
                        "font": self.font_size,
                        "space": self.spacing_scale,
                    },
                    fps=self.video_fps,
                    clf_name=self.clf_name,
                    bout_total_cnt=len(clf_bouts),
                    video_meta_data=self.video_info,
                    p_data=self.data_df[self.p_col].astype(np.float32),
                    clf_data=self.data_df[self.clf_name].astype(np.float32),
                    highlight_clr=self.highlight_clr,
                    text_clr=self.text_clr,
                )

                for cnt, result in enumerate(
                    pool.imap(
                        constants, clip_data, chunksize=self.multiprocess_chunksize
                    )
                ):
                    print(f"Bout {cnt+1} complete...")
                pool.terminate()
                pool.join()

            if self.concat_video:
                print(f"Joining {file_name} multiprocessed video...")
                concat_video_save_path = os.path.join(
                    self.clf_validation_dir,
                    f"{self.clf_name}_{file_name}_all_events.mp4",
                )
                file_paths = list(clip_data[:, 3])
                concatenate_videos_in_folder(
                    in_folder=self.clf_validation_dir,
                    save_path=concat_video_save_path,
                    substring=None,
                    file_paths=file_paths,
                    remove_splits=False,
                )
            if not self.clips:
                remove_files(file_paths=list(clip_data[:, 3]))
            video_timer.stop_timer()
            stdout_success(
                msg=f"Validation clips for video {file_name} complete!",
                elapsed_time=video_timer.elapsed_time_str,
            )

        self.timer.stop_timer()
        stdout_success(
            msg=f"All video clips complete and saved in {self.clf_validation_dir}!",
            elapsed_time=self.timer.elapsed_time_str,
        )


# test = ClassifierValidationClipsMultiprocess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                              window=1,
#                                              clf_name='Attack',
#                                              clips=True,
#                                              concat_video=True,
#                                              highlight_clr=(255, 0, 0),
#                                              video_speed=0.5,
#                                              text_clr=(0, 0, 255),
#                                              data_paths=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv',
#                                                          '/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_2.csv'])
# test.run()

# test = ClassifierValidationClips(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini',
#                                  window=1,
#                                  clf_name='Attack',
#                                  clips=False,
#                                  concat_video=True,
#                                  text_clr=(0, 0, 255))
# test.create_clips()
